# File for generating vascular networks for
# cube, cylinder, heart, and gyrus

import svcco
import pyvista as pv
from time import perf_counter
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from itertools import chain
from tqdm import trange

###########################################
# Code for Building Surfaces
###########################################
q = 4
resolution = 120


cu = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(5)
cube = svcco.surface()
cube.set_data(cu.points,cu.point_normals)
cube.solve()
cube.build(q=q,resolution=resolution)
print('cube constructed')

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()
heart.build(q=4,resolution=120,k=2,buffer=5)
print('heart constructed')


disk = pv.Disc(inner=2.8,outer=4,r_res=20,c_res=100)
cyl  = disk.extrude([0,0,2],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
cylinder = svcco.surface()
cylinder.set_data(cyl.points,cyl.point_normals)
cylinder.solve()
cylinder.build(q=q,resolution=resolution)
print('cylinder constructed')

left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"
gyrus_no_scale = pv.read(left_gyrus)
sf = (heart.volume/gyrus_no_scale.volume)**(1/3)
gyrus_scaled = gyrus_no_scale.scale([sf,sf,sf])
left_gyrus_scaled = "left_gyrus_scaled.vtp"
gyrus_scaled.save(left_gyrus_scaled)
gyrus = svcco.surface()
gyrus.load(left_gyrus_scaled)
gyrus.solve()
gyrus.build(q=q,resolution=resolution,buffer=5)
print('gyrus constructed')

############################################
# Generate Vessel Networks for Domains
############################################

vessel_number = 1000

cube_tree = svcco.tree()
cube_tree.set_boundary(cube)
cube_tree.convex = True
cube_tree.set_root()
cube_tree.n_add(vessel_number)

cube_tree_models = cube_tree.show(show=False)
cube_tree_plot   = pv.Plotter()
cube_tree_plot.add_mesh(cube_tree.boundary.pv_polydata,opacity=0.25,color='red')
for model in cube_tree_models:
    tmp = model.GetOutput()
    data  = pv.PolyData(tmp)
    _ = cube_tree_plot.add_mesh(data,color='red')
cube_tree_plot.set_background(color='white')
cube_tree_plot.camera_position = 'yz'
cube_tree_plot.camera.azimuth = 45
cube_tree_plot.save_graphic('cube_tree_vessels{}_q{}_res{}.svg'.format(vessel_number,q,resolution))
path = cube_tree_plot.generate_orbital_path(n_points=36,shift=cube.pv_polydata.length)
cube_tree_plot.open_gif("cube_orbit.gif")
cube_tree_plot.orbit_on_path(path,write_frames=True)
# Export SVG of Tree and surface

cylinder_tree = svcco.tree()
cylinder_tree.set_boundary(cylinder)
cylinder_tree.set_root()
cylinder_tree.n_add(vessel_number)

cylinder_tree_models = cylinder_tree.show(show=False)
cylinder_tree_plot   = pv.Plotter()
cylinder_tree_plot.add_mesh(cylinder_tree.boundary.pv_polydata,opacity=0.25,color='red')
for model in cylinder_tree_models:
    tmp = model.GetOutput()
    data = pv.PolyData(tmp)
    _ = cylinder_tree_plot.add_mesh(data,color='red')
cylinder_tree_plot.set_background(color='white')
cylinder_tree_plot.camera_position = 'yz'
cylinder_tree_plot.camera.roll = 0
cylinder_tree_plot.elevation = 25
cylinder_tree_plot.camera.azimuth = 140
cylinder_tree_plot.save_graphic('cylinder_tree_vessels{}_q{}_res{}.svg'.format(vessel_number,q,resolution))
path = cylinder_tree_plot.generate_orbital_path(n_points=36,shift=cylinder.pv_polydata.length)
cylinder_tree_plot.open_gif("cylinder_orbit.gif")
cylinder_tree_plot.orbit_on_path(path,write_frames=True)
# Export SVG of Tree and surface

heart_tree = svcco.tree()
heart_tree.set_boundary(heart)
heart_tree.set_root()
heart_tree.n_add(vessel_number)

# Export SVG of Tree and surface

heart_tree_models = heart_tree.show(show=False)
heart_tree_plot   = pv.Plotter()
heart_tree_plot.add_mesh(heart_tree.boundary.pv_polydata,opacity=0.25,color='red')
for model in heart_tree_models:
    tmp = model.GetOutput()
    data = pv.PolyData(tmp)
    _ = heart_tree_plot.add_mesh(data,color='red')
heart_tree_plot.set_background(color='white')
heart_tree_plot.camera_position = 'yz'
heart_tree_plot.camera.roll = 160
heart_tree_plot.save_graphic('heart_tree_vessels{}_q{}_res{}.svg'.format(vessel_number,q,resolution))
path = heart_tree_plot.generate_orbital_path(n_points=36,shift=heart.pv_polydata.length)
heart_tree_plot.open_gif("heart_orbit.gif")
heart_tree_plot.orbit_on_path(path,write_frames=True)


gyrus_tree = svcco.tree()
gyrus_tree.set_boundary(gyrus)
gyrus_tree.set_root()
gyrus_tree.n_add(vessel_number)

# Export SVG of Tree and surface

gyrus_tree_models = gyrus_tree.show(show=False)
gyrus_tree_plot   = pv.Plotter()
gyrus_tree_plot.add_mesh(gyrus_tree.boundary.pv_polydata,opacity=0.25,color='red')
for model in gyrus_tree_models:
    tmp = model.GetOutput()
    data = pv.PolyData(tmp)
    _ = gyrus_tree_plot.add_mesh(data,color='red')
gyrus_tree_plot.set_background(color='white')
gyrus_tree_plot.camera_position = 'yz'
gyrus_tree_plot.camera.azimuth = 45
gyrus_tree_plot.save_graphic('gyrus_tree_vessels{}_q{}_res{}.svg'.format(vessel_number,q,resolution))
path = gyrus_tree_plot.generate_orbital_path(n_points=36,shift=gyrus.pv_polydata.length)
gyrus_tree_plot.open_gif("gyrus_orbit.gif")
gyrus_tree_plot.orbit_on_path(path,write_frames=True)



# Competitive growth of Multiple Vascular Networks within a Shared Domain
import numpy as np
import os #remove before final
import time #not neccessary?
from ..branch_addition.check import *
from ..branch_addition.close import *
from ..branch_addition.local_func_v7 import *
from ..collision.sphere_proximity import *
from ..collision.collision import *
from ..branch_addition.add_bifurcation import *
from ..branch_addition.sample_triad import *
from ..branch_addition.triangle import * #might not need
from ..branch_addition.basis import *
#from scipy import interpolate
#from scipy.spatial import KDTree
from ..branch_addition.get_point import *
from copy import deepcopy
#from .geodesic import extract_surface,geodesic
#from scipy.sparse.csgraph import shortest_path
#from .add_geodesic_path import *
import pyvista as pv
import random
from tqdm import tqdm

def compete_add(forest,network_ids=-1,radius_buffer=0.05,exact=True):
    #Find a viable new point to add within
    all_network_ids = list(range(forest.number_of_networks))
    if network_ids == -1:
        network_ids = all_network_ids
    #points,_ = forest.boundary.pick(size=prefetch,homogeneous=True,replacement=False)
    #points = points.tolist()
    active_network_ids = deepcopy(network_ids)
    active_network_tree_ids = [list(range(forest.trees_per_network[net])) for net in all_network_ids]
    scale = len(network_ids)
    volume = deepcopy(forest.boundary.volume)
    forest.boundary.volume = forest.boundary.volume/scale
    #networks_copy = deepcopy(forest.networks)
    unused_networks = list(set(all_network_ids) - set(active_network_ids))
    while len(active_network_ids) > 0:
        for nid in active_network_ids:
            for njd in active_network_tree_ids[nid]:
                ## New
                if  forest.networks[nid][njd].data.shape[0] < 100:
                    if len(forest.networks[nid][njd].rng_points) <= 1:
                        forest.networks[nid][njd].rng_points,_ = forest.boundary.pick(size=min(len(forest.boundary.tet_verts),10000),homogeneous=True,replacement=False)
                        forest.networks[nid][njd].rng_points = forest.networks[nid][njd].rng_points.tolist()
                    for n in range(len(forest.networks[nid][njd].rng_points)):
                        pt = np.array(forest.networks[nid][njd].rng_points.pop(0))
                        if exact:
                            other_vessels,distances = close_exact(forest.networks[nid][njd].data,pt)
                        else:
                            other_vessels,distances = close(forest.networks[nid][njd].data,pt)
                        minimum_distance = min(distances)
                        retry = False
                        for idx in network_ids:
                            for jdx in list(range(forest.trees_per_network[idx])):
                                if idx == nid:
                                    continue
                                if forest.networks[nid][njd].data.shape[0] < 100:
                                    other_vessels,distances = close_exact(forest.networks[idx][jdx].data,pt)
                                else:
                                    other_vessels,distances = close(forest.networks[idx][jdx].data,pt)
                                if min(distances) < minimum_distance:
                                    retry = True
                                    break
                            if retry:
                                break
                        if retry:
                            continue
                        else:
                            forest.networks[nid][njd].rng_points.insert(-1,pt.tolist())
                ##
                #print(len(forest.networks[nid][njd].rng_points))
                #print("Check: {}".format(forest.networks[nid][njd].rng_points[0]))
                vessel,data,sub_division_map,sub_division_index,threshold = forest.networks[nid][njd].add(-1,0,isforest=True,radius_buffer=radius_buffer)
                point = data[-2,3:6] #new terminal point
                #print(point)
                closest_network = nid
                _,closest_value = close_exact(forest.networks[nid][njd].data[vessel,:].reshape(1,forest.networks[nid][njd].data.shape[1]),point)
                escape = False
                for idx in network_ids:
                    for jdx in list(range(forest.trees_per_network[idx])):
                        if idx == nid:
                            continue
                        if exact:
                            other_vessels,distances = close_exact(forest.networks[idx][jdx].data,point)
                        else:
                            other_vessels,distances = close(forest.networks[idx][jdx].data,point)
                        minimum_distance = min(distances)
                        minimum_vessel_point = (forest.networks[idx][jdx].data[other_vessels[0],0:3] + forest.networks[idx][jdx].data[other_vessels[0],3:6])/2
                        if minimum_distance < closest_value:
                            vessel_within = True
                            if not forest.convex:
                                subdivisions = 5
                                for sub in range(1,2*subdivisions):
                                    tmp = point*(sub/(2*subdivisions)) + minimum_vessel_point*(1-sub/(2*subdivisions))
                                    if not forest.boundary.within(tmp[0],tmp[1],tmp[2],2):
                                        vessel_within = False
                                        break
                            if not vessel_within:
                                continue
                            else:
                                escape = True
                                break
                    if escape:
                        break
                if escape:
                    #forest.networks[nid][njd] = deepcopy(networks_copy[nid][njd])
                    continue
                new_vessels = np.vstack((data[vessel,:],data[-2,:],data[-1,:]))
                repeat = False
                check_networks = unused_networks + [nid]
                for nikd in range(len(unused_networks)):
                    if nikd == nid:
                        check_trees = list(range(len(forest.networks[nikd])))
                        check_trees.remove(njd)
                    else:
                        check_trees = list(range(len(forest.networks[nikd])))
                    for njkd in check_trees:
                        if not forest.networks[nikd][njkd].collision_free(new_vessels,radius_buffer=radius_buffer):
                            repeat = True
                            break
                    if repeat:
                        break
                if repeat:
                    #forest.networks[nid][njd] = deepcopy(networks_copy[nid][njd])
                    continue
                else:
                    forest.networks[nid][njd].data = data
                    forest.networks[nid][njd].parameters['edge_num'] += 2
                    forest.networks[nid][njd].sub_division_map = sub_division_map
                    forest.networks[nid][njd].sub_division_index = sub_division_index
                    #print("List {}: {},  Id: {}".format(nid,active_network_tree_ids[nid],njd))
                    active_network_tree_ids[nid].remove(njd)
                    if len(active_network_tree_ids[nid]) == 0:
                        active_network_ids.remove(nid)
                    #print("Active Networks: {}".format(active_network_ids))
    forest.boundary.volume = volume



"""
File for loading spline functions.
"""
import tkinter as tk
from tkinter.filedialog import askopenfilename
import scipy as sp
import pickle

def load_splines(filename=None):
    """
    This file loads Spline functions that have
    been pickled into a binary file (*.pkl file)

    Parameters
    ----------
              filename: str, optional
                       the path to the spline file.
                       if this is not provided then
                       the function will prompt the
                       user to fetch the file with a
                       dialog window popup.

    Returns
    -------
              splines: List
                     a list of spline functions. each
                     list item is a different N-dimensional
                     spline which can be evaluated along the
                     interval [0,1]. Vessel data evaluations
                     are given in xyzr format. This means
                     the x-coordinate, y-coordinate, z-coordinate,
                     and radius value are returned in this
                     order.

    Example:

    >>> splines = load_splines()
    >>> splines[0](0)     # we are evaluating the first spline at zero
    [array(0.02465846), array(-1.3615156), array(1.86000645), array(0.06768623)]
    >>>
    >>> x,y,z,r = splines[0](0) # we can store the data into the relevant variables
    >>> print(x)
    array(0.02465846)
    >>>print(y)
    array(-1.3615156)
    >>>print(z)
    array(1.86000645)
    >>>print(r)
    array(0.06768623)
    
    """
    filename = askopenfilename()
    file     = open(filename,'rb')
    splines  = pickle.load(file)
    return splines



flow="""if not terminating:
    flow_file_3d = open('{}'+os.sep+'{}'+'_3d.flow','w+')
    flow_file_rom = open('{}'+os.sep+'{}'+'_rom.flow','w+')
    flow_file_lines_3d = []
    flow_file_lines_rom = []
    times = {}
    flows = {}
    for t,f in zip(times,flows):
        flow_file_lines_3d.append(str(t)+' -'+str(f)+'\\n')
        flow_file_lines_rom.append(str(t)+' '+str(f)+'\\n')
    flow_file_3d.writelines(flow_file_lines_3d)
    flow_file_rom.writelines(flow_file_lines_rom)
    flow_file_3d.close()
    flow_file_rom.close()
"""



import svcco
import numpy as np
import matplotlib.pyplot as plt
import pyvista as pv

q = 4
resolution = 120


cu = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(5)
cube = svcco.surface()
cube.set_data(cu.points,cu.point_normals)
cube.solve()
cube.build(q=q,resolution=resolution)
print('cube constructed')

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()
heart.build(q=q,resolution=resolution,buffer=5)
print('heart constructed')


disk = pv.Disc(inner=2.8,outer=4,r_res=20,c_res=100)
cyl  = disk.extrude([0,0,2],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
cylinder = svcco.surface()
cylinder.set_data(cyl.points,cyl.point_normals)
cylinder.solve()
cylinder.build(q=q,resolution=resolution)
print('cylinder constructed')

left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"
gyrus_no_scale = pv.read(left_gyrus)
sf = (heart.volume/gyrus_no_scale.volume)**(1/3)
gyrus_scaled = gyrus_no_scale.scale([sf,sf,sf])
left_gyrus_scaled = "left_gyrus_scaled.vtp"
gyrus_scaled.save(left_gyrus_scaled)
gyrus = svcco.surface()
gyrus.load(left_gyrus_scaled)
gyrus.solve()
gyrus.build(q=q,resolution=resolution,buffer=5)
print('gyrus constructed')


###########################################
# Test function
###########################################

def test(surf_object,size=1000):
    t = svcco.tree()
    t.set_boundary(surf_object)
    delaunay = surf_object.pv_polydata.delaunay_3d()
    convexity = surf_object.volume/delaunay.volume
    if convexity > 0.95:
        t.convex = True
    t.set_root()
    t.n_add(size)

    w = svcco.utils.fluid_analysis.wss.wss(t)
    r = t.data[:,21]*10 # in millimeters
    l = t.data[:,20]*10 # in millimeters

    a = []
    for i in range(t.data.shape[0]):
        if t.data[i,15] > 0:
            dau_1 = int(t.data[i,15].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_1,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_1,12:15]))
            if dot > 1:
                dot = 1
            ang_1 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_1)
        if t.data[i,16] > 0:
            dau_2 = int(t.data[i,16].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_2,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_2,12:15]))
            if dot > 1:
                dot = 1
            ang_2 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_2)
    a = np.array(a)

    branches = svcco.sv_interface.get_sv_data.get_branches(t.data)
    torts = []
    for branch in branches:
        tort = 0
        for idx in range(len(branch)-1):
            dot = np.dot(t.data[branch[idx],12:15],t.data[branch[idx+1],12:15].T)
            dot = dot/(np.linalg.norm(t.data[branch[idx],12:15])*np.linalg.norm(t.data[branch[idx+1],12:15]))
            if dot > 1:
                dot = 1
            ang = 180*(np.arccos(dot)/np.pi)
            tort += ang
        torts.append(tort)
    torts = np.array(torts)
    return w,r,l,a,torts

def test_heart(size=1000):
    t = svcco.tree()
    t.set_boundary(heart)
    t.set_root()
    t.n_add(size)

    w = svcco.utils.fluid_analysis.wss.wss(t)
    r = t.data[:,21]*10 #should be microns
    l = t.data[:,20]*10

    a = []
    for i in range(t.data.shape[0]):
        if t.data[i,15] > 0:
            dau_1 = int(t.data[i,15].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_1,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_1,12:15]))
            if dot > 1:
                dot = 1
            ang_1 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_1)
        if t.data[i,16] > 0:
            dau_2 = int(t.data[i,16].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_2,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_2,12:15]))
            if dot > 1:
                dot = 1
            ang_2 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_2)
    a = np.array(a)

    branches = svcco.sv_interface.get_sv_data.get_branches(t.data)
    torts = []
    for branch in branches:
        tort = 0
        for idx in range(len(branch)-1):
            dot = np.dot(t.data[branch[idx],12:15],t.data[branch[idx+1],12:15].T)
            dot = dot/(np.linalg.norm(t.data[branch[idx],12:15])*np.linalg.norm(t.data[branch[idx+1],12:15]))
            if dot > 1:
                dot = 1
            ang = 180*(np.arccos(dot)/np.pi)
            tort += ang
        torts.append(tort)
    torts = np.array(torts)
    return w,r,l,a,torts

def test_sphere(size=1000):
    t = svcco.tree()
    t.set_boundary(sph)
    t.convex = True
    t.set_root()
    t.n_add(size)

    w = svcco.utils.fluid_analysis.wss.wss(t)
    r = t.data[:,21]*10 #should be microns
    l = t.data[:,20]*10

    a = []
    for i in range(t.data.shape[0]):
        if t.data[i,15] > 0:
            dau_1 = int(t.data[i,15].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_1,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_1,12:15]))
            if dot > 1:
                dot = 1
            ang_1 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_1)
        if t.data[i,16] > 0:
            dau_2 = int(t.data[i,16].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_2,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_2,12:15]))
            if dot > 1:
                dot = 1
            ang_2 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_2)
    a = np.array(a)

    branches = svcco.sv_interface.get_sv_data.get_branches(t.data)
    torts = []
    for branch in branches:
        tort = 0
        for idx in range(len(branch)-1):
            dot = np.dot(t.data[branch[idx],12:15],t.data[branch[idx+1],12:15].T)
            dot = dot/(np.linalg.norm(t.data[branch[idx],12:15])*np.linalg.norm(t.data[branch[idx+1],12:15]))
            if dot > 1:
                dot = 1
            ang = 180*(np.arccos(dot)/np.pi)
            tort += ang
        torts.append(tort)
    torts = np.array(torts)
    return w,r,l,a,torts

def test_gyrus(size=1000):
    t = svcco.tree()
    t.set_boundary(gyrus)
    t.set_root()
    t.n_add(size)

    w = svcco.utils.fluid_analysis.wss.wss(t)
    r = t.data[:,21]*10 #should be microns
    l = t.data[:,20]*10

    a = []
    for i in range(t.data.shape[0]):
        if t.data[i,15] > 0:
            dau_1 = int(t.data[i,15].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_1,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_1,12:15]))
            if dot > 1:
                dot = 1
            ang_1 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_1)
        if t.data[i,16] > 0:
            dau_2 = int(t.data[i,16].item())
            dot   = np.dot(t.data[i,12:15],t.data[dau_2,12:15].T)
            dot   = dot/(np.linalg.norm(t.data[i,12:15])*np.linalg.norm(t.data[dau_2,12:15]))
            if dot > 1:
                dot = 1
            ang_2 = 180*(np.arccos(dot)/np.pi)
            a.append(ang_2)
    a = np.array(a)

    branches = svcco.sv_interface.get_sv_data.get_branches(t.data)
    torts = []
    for branch in branches:
        tort = 0
        for idx in range(len(branch)-1):
            dot = np.dot(t.data[branch[idx],12:15],t.data[branch[idx+1],12:15].T)
            dot = dot/(np.linalg.norm(t.data[branch[idx],12:15])*np.linalg.norm(t.data[branch[idx+1],12:15]))
            if dot > 1:
                dot = 1
            ang = 180*(np.arccos(dot)/np.pi)
            tort += ang
        torts.append(tort)
    torts = np.array(torts)
    return w,r,l,a,torts


def results(test_list,size=1000,iter=10,bins=50,range_w=(0,1500),range_r=(0,1),range_l=(0,4),range_a = (0,90),range_t=(0,4*180)):
    WSS_ALL = []
    RAD_ALL = []
    LEN_ALL = []
    ANG_ALL = []
    TORT_ALL = []
    WSS_ALL_MEAN = []
    RAD_ALL_MEAN = []
    LEN_ALL_MEAN = []
    ANG_ALL_MEAN = []
    TORT_ALL_MEAN = []
    WSS_ALL_STD  = []
    RAD_ALL_STD  = []
    LEN_ALL_STD  = []
    ANG_ALL_STD  = []
    TORT_ALL_STD = []
    WSS_WIDTH_ALL= []
    RAD_WIDTH_ALL= []
    LEN_WIDTH_ALL= []
    ANG_WIDTH_ALL= []
    TORT_WIDTH_ALL=[]
    WSS_CENTER_ALL = []
    RAD_CENTER_ALL = []
    LEN_CENTER_ALL = []
    ANG_CENTER_ALL = []
    TORT_CENTER_ALL = []
    for i in range(len(test_list)):
        WSS = []
        RAD = []
        LEN = []
        ANG = []
        TORT = []
        for j in range(iter):
            w,r,l,a,torts = test(test_list[i],size=size)
            freq_w,edges_w = np.histogram(w,bins=bins,range=range_w)
            freq_r,edges_r = np.histogram(r,bins=bins,range=range_r)
            freq_l,edges_l = np.histogram(l,bins=bins,range=range_l)
            freq_a,edges_a = np.histogram(a,bins=bins,range=range_a)
            freq_t,edges_t = np.histogram(torts,bins=bins,range=range_t)
            WSS.append(freq_w.tolist())
            RAD.append(freq_r.tolist())
            LEN.append(freq_l.tolist())
            ANG.append(freq_a.tolist())
            TORT.append(freq_t.tolist())
        WSS_ALL.append(WSS)
        RAD_ALL.append(RAD)
        LEN_ALL.append(LEN)
        ANG_ALL.append(ANG)
        TORT_ALL.append(TORT)
        WSS = np.array(WSS)
        RAD = np.array(RAD)
        LEN = np.array(LEN)
        ANG = np.array(ANG)
        #ANG = ANG[np.isfinite(ANG)]
        TORT = np.array(TORT)
        #TORT = TORT[np.isfinite(TORT)]
        WSS_mean = np.mean(WSS,axis=0)
        WSS_ALL_MEAN.append(WSS_mean)
        WSS_std  = np.std(WSS,axis=0)
        WSS_ALL_STD.append(WSS_std)
        RAD_mean = np.mean(RAD,axis=0)
        RAD_ALL_MEAN.append(RAD_mean)
        RAD_std  = np.std(RAD,axis=0)
        RAD_ALL_STD.append(RAD_std)
        LEN_mean = np.mean(LEN,axis=0)
        LEN_ALL_MEAN.append(LEN_mean)
        LEN_std  = np.std(LEN,axis=0)
        LEN_ALL_STD.append(LEN_std)
        ANG_mean = np.mean(ANG,axis=0)
        ANG_ALL_MEAN.append(ANG_mean)
        ANG_std  = np.std(ANG,axis=0)
        ANG_ALL_STD.append(ANG_std)
        TORT_mean = np.mean(TORT,axis=0)
        TORT_ALL_MEAN.append(TORT_mean)
        TORT_std  = np.std(TORT,axis=0)
        TORT_ALL_STD.append(TORT_std)

        width_w = (edges_w[1]-edges_w[0])
        WSS_WIDTH_ALL.append(width_w)
        width_r = (edges_r[1]-edges_r[0])
        RAD_WIDTH_ALL.append(width_r)
        width_l = (edges_l[1]-edges_l[0])
        LEN_WIDTH_ALL.append(width_l)
        width_a = (edges_a[1]-edges_a[0])
        ANG_WIDTH_ALL.append(width_a)
        width_t = (edges_t[1]-edges_t[0])
        TORT_WIDTH_ALL.append(width_t)
        centers_w = 0.5*(edges_w[1:]+edges_w[:-1])
        WSS_CENTER_ALL.append(centers_w)
        centers_r = 0.5*(edges_r[1:]+edges_r[:-1])
        RAD_CENTER_ALL.append(centers_r)
        centers_l = 0.5*(edges_l[1:]+edges_l[:-1])
        LEN_CENTER_ALL.append(centers_l)
        centers_a = 0.5*(edges_a[1:]+edges_a[:-1])
        ANG_CENTER_ALL.append(centers_a)
        centers_t = 0.5*(edges_t[1:]+edges_t[:-1])
        TORT_CENTER_ALL.append(centers_t)
    fig,ax = plt.subplots(4,len(test_list))
    for i in range(len(test_list)):
        centers_w = WSS_CENTER_ALL[i]
        WSS_mean  = WSS_ALL_MEAN[i]
        width_w   = WSS_WIDTH_ALL[i]
        WSS_std   = WSS_ALL_STD[i]

        centers_r = RAD_CENTER_ALL[i]
        RAD_mean  = RAD_ALL_MEAN[i]
        width_r   = RAD_WIDTH_ALL[i]
        RAD_std   = RAD_ALL_STD[i]

        centers_l = LEN_CENTER_ALL[i]
        LEN_mean  = LEN_ALL_MEAN[i]
        width_l   = LEN_WIDTH_ALL[i]
        LEN_std   = LEN_ALL_STD[i]

        centers_a = ANG_CENTER_ALL[i]
        ANG_mean  = ANG_ALL_MEAN[i]
        width_a   = ANG_WIDTH_ALL[i]
        ANG_std   = ANG_ALL_STD[i]

        centers_t = TORT_CENTER_ALL[i]
        TORT_mean = TORT_ALL_MEAN[i]
        width_t   = TORT_WIDTH_ALL[i]
        TORT_std  = TORT_ALL_STD[i]
        #ax[0][0].bar(centers_w,WSS_mean,width=width_w,yerr=WSS_std)
        if len(test_list) > 1:
            ax[0][i].bar(centers_r,RAD_mean,width=width_r,yerr=RAD_std)
            ax[1][i].bar(centers_l,LEN_mean,width=width_l,yerr=LEN_std)
            ax[2][i].bar(centers_a,ANG_mean,width=width_a,yerr=ANG_std)
            ax[3][i].bar(centers_t,TORT_mean,width=width_t,yerr=TORT_std)
            #for pos,y,err in zip(centers_w,WSS_mean,WSS_std):
            #    ax[0][0].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_r,RAD_mean,RAD_std):
                ax[0][i].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_l,LEN_mean,LEN_std):
                ax[1][i].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_a,ANG_mean,ANG_std):
                ax[2][i].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_t,TORT_mean,TORT_std):
                ax[3][i].errorbar(pos,y,err,capsize=2,color='black')
            ax[0][i].set_xlim([0,None])
            ax[0][i].set_ylim([0,500])

            ax[1][i].set_xlim([0,None])
            ax[1][i].set_ylim([0,100])

            ax[2][i].set_xlim([0,90])
            ax[2][i].set_ylim([0,125])

            ax[3][i].set_xlim([0,None])
            ax[3][i].set_ylim([0,350])
        else:
            ax[0].bar(centers_r,RAD_mean,width=width_r,yerr=RAD_std)
            ax[1].bar(centers_l,LEN_mean,width=width_l,yerr=LEN_std)
            ax[2].bar(centers_a,ANG_mean,width=width_a,yerr=ANG_std)
            ax[3].bar(centers_t,TORT_mean,width=width_t,yerr=TORT_std)
            #for pos,y,err in zip(centers_w,WSS_mean,WSS_std):
            #    ax[0][0].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_r,RAD_mean,RAD_std):
                ax[0].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_l,LEN_mean,LEN_std):
                ax[1].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_a,ANG_mean,ANG_std):
                ax[2].errorbar(pos,y,err,capsize=2,color='black')
            for pos,y,err in zip(centers_t,TORT_mean,TORT_std):
                ax[3].errorbar(pos,y,err,capsize=2,color='black')
            ax[0].set_xlim([0,None])
            ax[0].set_ylim([0,None])

            ax[1].set_xlim([0,None])
            ax[1].set_ylim([0,None])

            ax[2].set_xlim([0,90])
            ax[2].set_ylim([0,None])

            ax[3].set_xlim([0,None])
            ax[3].set_ylim([0,None])
    if len(test_list) > 1:
        ax[0][0].set_ylabel('Number of Vessels')
        ax[1][0].set_ylabel('Number of Vessels')
        ax[2][0].set_ylabel('Number of Vessels')
        ax[3][0].set_ylabel('Number of Vessels')
        for i in range(len(test_list)):
            ax[0][i].set_xlabel('Vessel Radius (mm)')
            ax[1][i].set_xlabel('Vessel Length (mm)')
            ax[2][i].set_xlabel('Parent-daughter Angles')
            ax[3][i].set_xlabel('Tortuosity')
    else:
        ax[0].set_ylabel('Number of Vessels')
        ax[1].set_ylabel('Number of Vessels')
        ax[2].set_ylabel('Number of Vessels')
        ax[3].set_ylabel('Number of Vessels')
        ax[0].set_xlabel('Vessel Radius (mm)')
        ax[1].set_xlabel('Vessel Length (mm)')
        ax[2].set_xlabel('Parent-daughter Angles')
        ax[3].set_xlabel('Tortuosity')
    fig.savefig('morphometry_surfaces-{}_num_vessels-{}_iter-{}_num_bins-{}.svg'.format(len(test_list),size,iter,bins),format='svg')
    return fig,ax

f,a = results([cube,cylinder,heart,gyrus])



import numba as nb
import numpy as np

@nb.jit(nopython=True)
def get_path(D,Pr,i,j):
    path = [j]
    length = 0
    while Pr[i,k] != -9999:
        path.append(Pr[i,k])
        length += D[i,k]
        k = Pr[i,k]
    return path[::-1],length



from __future__ import print_function
import numba as nb
import vtk
from vtk.util import numpy_support
from scipy import optimize
from .visualize.visualize import *
from .sampling import *
from concurrent.futures import ProcessPoolExecutor as PPE
from concurrent.futures import as_completed
from functools import partial
from multiprocessing import Pool, RLock, freeze_support
from threading import RLock as TRLock
from .derivative import *
from tqdm.auto import tqdm, trange
from functools import partial
from .solver.solver import solver
from ..branch_addition.basis import tangent_basis
from pickle import dumps,loads
from scipy.spatial import cKDTree
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path
from scipy.interpolate import splprep,splev
#from gekko import GEKKO
import pymeshfix
import marshal, types
import tkinter as tk
from tkinter.filedialog import askopenfilename

import pyvista as pv
import tetgen
import time
import os

import imp
import sys

from .core.m_matrix import M
from .core.n_matrix import N
from .core.a_matrix import A
from .core.h_matrix import H
from .load import load3d, load3d_pv
from .visualize.visualize import show_mesh

class patch:
    def __init__(self,points,normals):
        self.ndim = points.shape[0]
        self.ddim = points.shape[1]
        self.points = points
        self.normals = normals
        start = time.perf_counter()
        self.A_inv,self.K00,self.K01,self.K11 = A(points)
        self.time = time.perf_counter() - start
        self.H_0 = H(self.K00,self.K01,self.K11,0)

    def solve(self,local_method='L-BFGS-B',regularize=False,
              local_verbosity=False,variational=False,
              solver_method='Bounded',solver_verbosity=False):
        if regularize:
            s = solver(self.points,self.normals)
            a,b = s.solve(local_verbosity=local_verbosity,local_method=local_method,
                          variational=variational,solver_method=solver_method,
                          solver_verbosity=solver_verbosity)
            self.a = a
            self.b = b
        else:
            g = np.zeros(self.ndim*3)
            M_inv = self.A_inv[:self.ndim*4,:self.ndim*4]
            N_inv = self.A_inv[:self.ndim*4,self.ndim*4:]
            for i in range(self.ndim):
                g[i] = self.normals[i,0]
                g[i+self.ndim] = self.normals[i,1]
                g[i+2*self.ndim] = self.normals[i,2]
            self.g = g
            s = np.zeros(self.ndim)
            l_side = np.zeros(M_inv.shape[1])
            l_side[:len(s)] = s
            l_side[len(s):(len(s)+len(g))] = g
            a = np.matmul(M_inv,l_side)
            b = np.matmul(N_inv.T,l_side)
            self.a = a
            self.b = b

class surface:
    def __init__(self):
        self.subtracted_volumes = []
        self.added_volumes = []
        self.polydata = None
        pass
    def load(self,*args,**kwargs):
        subdivisions = kwargs.get('subdivisions',0)
        max_points   = kwargs.get('max_points',10000)
        if len(args) == 0:
            root = tk.Tk()
            root.withdraw()
            filename = askopenfilename()
            root.update()
        else:
            filename = args[0]
        points,normals,polydata = load3d_pv(filename,subdivisions,max_points=max_points)
        self.set_data(points,normals)
        self.polydata = polydata
    def set_data(self,*args,workers=1,local_min=10,local_max=20,l=0.5,PCA_samplesize=25):
        if len(args) == 0:
            root = tk.Tk()
            root.withdraw()
            point_filename = askopenfilename(filetypes=[('CSV Files','*.csv')])
            root.update()
            points = np.genfromtxt(point_filename,delimiter=',')
            get_normals = input('Imput normals as well? y/n \n')
            if get_normals == 'y':
                root = tk.Tk()
                root.withdraw()
                normal_filename = askopenfilename(filetypes=[('CSV Files','*.csv')])
                root.update()
                normals = np.genfromtxt(normal_filename,delimiter=',')
            else:
                normals = None
        elif len(args) == 1:
            points  = args[0]
            normals = None
        elif len(args) == 2:
            points  = args[0]
            normals = args[1]
        self.points = points
        self.ddim = points.shape[1]
        self.dim_range = [0]*self.ddim*2
        for i in range(self.ddim):
        	self.dim_range[i*2]   = min(points[:,i])
        	self.dim_range[i*2+1] = max(points[:,i])
        self.local_min = local_min
        self.local_max = local_max
        self.l = l
        self.workers = workers
        if np.any(normals) is None:
            point_cloud_vtk_array  = numpy_support.numpy_to_vtk(points,deep=True)
            point_cloud_vtk_points = vtk.vtkPoints()
            point_cloud_vtk_points.SetData(point_cloud_vtk_array)
            point_cloud_polydata   = vtk.vtkPolyData()
            point_cloud_polydata.SetPoints(point_cloud_vtk_points)
            PCA_normals            = vtk.vtkPCANormalEstimation()
            PCA_normals.SetInputData(point_cloud_polydata)
            PCA_normals.SetSampleSize(PCA_samplesize)
            PCA_normals.SetNormalOrientationToGraphTraversal()
            PCA_normals.Update()
            PCA_polydata = PCA_normals.GetOutput()
            PCA_vtk_pts  = PCA_polydata.GetPointData()
            vtk_normals  = PCA_vtk_pts.GetNormals()
            self.normals = numpy_support.vtk_to_numpy(vtk_normals)
        else:
            self.normals = normals

    @staticmethod
    def patch_constructor_progress(tuple_data, auto_position=False,
                                   write_safe=False, blocking=True,
                                   progress=True,regularize=False,
                                   local_verbosity=False,local_method='L-BFGS-B',
                                   variational=False,solver_method='Bounded',
                                   solver_verbosity=True):
        points = tuple_data[0]
        normals = tuple_data[1]
        chunk = tuple_data[2]
        patches = []
        total = len(points)
        text = "Processing Chunk {0}  : ".format(chunk)
        for i in trange(total, desc=text, leave=False, disable=not progress,
                        lock_args=None if blocking else (False,),
                        position=None if auto_position else chunk):
            p_tmp = patch(points[i],normals[i])
            p_tmp.solve(regularize=regularize,local_verbosity=local_verbosity,
                        local_method=local_method,variational=varitaional,
                        solver_method=solver_method,solver_verbosity=solver_verbosity)
            patches.append(p_tmp)
        return patches

    @staticmethod
    def patch_constructor_no_progress(tuple_data, auto_position=False,
                                      write_safe=False, blocking=True,
                                      progress=True,regularize=False,
                                      local_verbosity=False,local_method='L-BFGS-B',
                                      variational=False,solver_method='Bounded',
                                      solver_verbosity=True):
        points = tuple_data[0]
        normals = tuple_data[1]
        chunk = tuple_data[2]
        patches = []
        total = len(points)
        for i in range(total):
            p_tmp = patch(points[i],normals[i])
            p_tmp.solve(regularize=regularize,local_verbosity=local_verbosity,
                        local_method=local_method,variational=variational,
                        solver_method=solver_method,solver_verbosity=solver_verbosity)
            patches.append(p_tmp)
        return patches

    def parallel_build(self,chunks,show_individual=False,PU=True):
        patches = []
        if show_individual and PU:
            freeze_support()
            tqdm.set_lock(RLock())
            executor = PPE(max_workers=len(chunks),initializer=tqdm.set_lock,initargs=(tqdm.get_lock(),))
            patch_function = partial(self.patch_constructor_no_progress,regularize=self.__regularize__,
                                     local_verbosity=self.__local_verbosity__,local_method=self.__local_method__,
                                     variational=self.__variational__,solver_method=self.__solver_method__,
                                     solver_verbosity=self.__solver_verbosity__)
            futures = {executor.submit(patch_function,chunk): chunk for chunk in chunks}
        elif self.workers == 1 and PU:
            patch_function = partial(self.patch_constructor_no_progress,regularize=self.__regularize__,
                                     local_verbosity=self.__local_verbosity__,local_method=self.__local_method__,
                                     variational=self.__variational__,solver_method=self.__solver_method__,
                                     solver_verbosity=self.__solver_verbosity__)
            futures = patch_function(chunks[0])
        elif PU:
            executor = PPE(max_workers=len(chunks))
            patch_function = partial(self.patch_constructor_no_progress,regularize=self.__regularize__,
                                     local_verbosity=self.__local_verbosity__,local_method=self.__local_method__,
                                     variational=self.__variational__,solver_method=self.__solver_method__,
                                     solver_verbosity=self.__solver_verbosity__)
            futures = {executor.submit(patch_function,chunk): chunk for chunk in chunks}
        else:
            global_patch = patch(self.points,self.normals)
            global_patch.solve(local_method=self.__local_method__,regularize=self.__regularize__,
                               local_verbosity=self.__local_verbosity__,variational=self.__variational__,
                               solver_method=self.__solver_method__,solver_verbosity=self.__solver_verbosity__)
        total_patches = []
        if show_individual and PU:
            for future in as_completed(futures):
                 total_patches.extend(future.result())
        elif self.workers == 1 and PU:
            total_patches = futures
        elif PU:
            for future in as_completed(futures):
                total_patches.extend(future.result())
        else:
            total_patches.append(global_patch)
        if self.workers != 1:
            executor.shutdown()
        return total_patches

    def solve(self,angle=0,PU=True,show_individual=False,regularize=False,
              local_verbosity=False,local_method='L-BFGS-B',variational=False,
              solver_method='Bounded',solver_verbosity=False,quiet=True):
        self.__regularize__ = regularize
        self.__local_verbosity__ = local_verbosity
        self.__local_method__ = local_method
        self.__variational__ = variational
        self.__solver_method__ = solver_method
        self.__solver_verbosity__= solver_verbosity
        if PU:
            patches,idxs,KDTree,centers = sampling(self.points,
                                           normals=self.normals,
                                           min_local_size=self.local_min,
                                           max_local_size=self.local_max,l=self.l,
                                           angle=angle,quiet=quiet)
            patches = []
            patch_centers = []
            patch_center_idx = []
            self.max_patch_size = 0
            chunksize = len(idxs)//self.workers
            number_of_chunks = len(idxs)//chunksize
            chunks = []
            for i in range(number_of_chunks):
                tmp_points = []
                tmp_normals = []
                if i < number_of_chunks-1:
                    if quiet:
                        for j in range(i*chunksize,i*chunksize+chunksize):
                            tmp_points.append(self.points[idxs[j]])
                            tmp_normals.append(self.normals[idxs[j]])
                            if len(idxs[j]) > self.max_patch_size:
                                self.max_patch_size = len(idxs[j])
                    else:
                        for j in tqdm(range(i*chunksize,i*chunksize+chunksize),desc='Building Chunk {}       '.format(i)):
                            tmp_points.append(self.points[idxs[j]])
                            tmp_normals.append(self.normals[idxs[j]])
                            if len(idxs[j]) > self.max_patch_size:
                                self.max_patch_size = len(idxs[j])
                    chunk_chunk = i
                    chunks.append([tmp_points,tmp_normals,chunk_chunk])
                else:
                    if i*chunksize >= len(idxs):
                        break
                    else:
                        if quiet:
                            for j in range(i*chunksize,len(idxs)):
                                tmp_points.append(self.points[idxs[j]])
                                tmp_normals.append(self.normals[idxs[j]])
                                if len(idxs[j]) > self.max_patch_size:
                                    self.max_patch_size = len(idxs[j])
                        else:
                            for j in tqdm(range(i*chunksize,len(idxs)),desc='Building Chunk {}       '.format(i)):
                                tmp_points.append(self.points[idxs[j]])
                                tmp_normals.append(self.normals[idxs[j]])
                                if len(idxs[j]) > self.max_patch_size:
                                    self.max_patch_size = len(idxs[j])
                        chunk_chunk = i
                        chunks.append([tmp_points,tmp_normals,chunk_chunk])
            _ = patch(chunks[0][0][0],chunks[0][1][0])
            self.patches = self.parallel_build(chunks,show_individual=show_individual)
        else:
            self.max_patch_size = self.points.shape[0]
            self.patches = self.parallel_build(None,show_individual=show_individual,PU=PU)

    def build(self,**kwargs):
        h          = kwargs.get('h',0)
        q          = kwargs.get('q',1)
        d_num      = kwargs.get('d_num',2)
        resolution = kwargs.get('resolution',20)
        k          = kwargs.get('k',2)
        level      = kwargs.get('level',0)
        visualize  = kwargs.get('visualize',False)
        buffer     = kwargs.get('buffer',1)
        verbose    = kwargs.get('verbose',False)
        self.x_range = [min(self.points[:,0]),max(self.points[:,0])]
        self.y_range = [min(self.points[:,1]),max(self.points[:,1])]
        self.z_range = [min(self.points[:,2]),max(self.points[:,2])]
        solution_matrix = []
        a_coef          = []
        b_coef          = []
        b_sol           = []
        n_points        = []
        patch_points    = []
        patch_x         = []
        patch_y         = []
        patch_z         = []
        patch_points_m  = []
        for patch in self.patches:
            solution_matrix_tmp = np.empty((self.max_patch_size,len(patch.b)))
            b_sol_tmp = np.zeros((self.max_patch_size,len(patch.b)))
            points = np.empty((self.max_patch_size,len(patch.b)-1))
            patch_points_m.append(patch.points)
            solution_matrix_tmp.fill(np.nan)
            points.fill(np.nan)
            points[:patch.ndim,:] = patch.points
            a_coef.append(patch.a)
            b_coef.append(patch.b)
            n_points.append(patch.ndim)
            patch_points.append(points)
            patch_x.append(patch.points[0,0])
            patch_y.append(patch.points[0,1])
            patch_z.append(patch.points[0,2])
            solution_matrix_tmp[:patch.ndim,0] = patch.a[:patch.ndim]
            b_sol_tmp[0,:] = patch.b
            for point in range(patch.ndim):
                solution_matrix_tmp[point,1] = patch.a[patch.ndim+point]
                solution_matrix_tmp[point,2] = patch.a[patch.ndim*2+point]
                solution_matrix_tmp[point,3] = patch.a[patch.ndim*3+point]
            solution_matrix.append(solution_matrix_tmp)
            b_sol.append(b_sol_tmp)
        self.solution_matrix = solution_matrix
        self.a_coef          = a_coef
        self.b_coef          = b_coef
        self.b_sol           = b_sol
        self.n_points        = n_points
        self.patch_points    = patch_points
        self.patch_x         = patch_x
        self.patch_y         = patch_y
        self.patch_z         = patch_z
        self.patch_points_m  = patch_points_m
        if len(self.patches) > 1:
            functions = []
            KDTree = spatial.cKDTree(np.array([patch_x,patch_y,patch_z]).T)
            self.patch_KDTree = KDTree
            preassembled_functions,function_strings = construct(d_num)
            foo = imp.new_module("foo")
            sys.modules["foo"] = foo
            pickled_DD = []
            DD = []
            exec("import numpy as np",foo.__dict__)
            for fss in function_strings:
                exec(fss,foo.__dict__)
            function_names = list(filter(lambda item: '__' not in item,dir(foo)))
            for fn in function_names:
                if fn == 'np':
                    continue
                exec("pickled_DD.append(dumps(partial(foo.{},KDTree=KDTree,patch_points".format(fn)+\
                     "=np.array(patch_points),b_coef=np.array(b_sol),h=h,q=q,sol_mat=np"+\
                     ".array(solution_matrix),patch_1=np.array(patch_x),patch_2=np.arra"+\
                     "y(patch_y),patch_3=np.array(patch_z))))")
                exec("DD.append(partial(foo.{},KDTree=KDTree,patch_points".format(fn)+\
                     "=np.array(patch_points),b_coef=np.array(b_sol),h=h,q=q,sol_mat=np"+\
                     ".array(solution_matrix),patch_1=np.array(patch_x),patch_2=np.arra"+\
                     "y(patch_y),patch_3=np.array(patch_z)))")
            func_marching = partial(function_marching,patch_points=patch_points_m,
                                    a_coef=a_coef,b_coef=b_coef,h=h,q=q,
                                    patch_x=np.array(patch_x),patch_y=np.array(patch_y),
                                    patch_z=np.array(patch_z))
            self.function_marching = func_marching
        else:
            functions = []
            preassembled_functions,function_strings = construct_global(d_num)
            foo = imp.new_module("foo")
            sys.modules["foo"] = foo
            pickled_DD = []
            DD = []
            exec("import numpy as np",foo.__dict__)
            for fss in function_strings:
                exec(fss,foo.__dict__)
            function_names = list(filter(lambda item: '__' not in item,dir(foo)))
            for fn in function_names:
                if fn == 'np':
                    continue
                exec("pickled_DD.append(dumps(partial(foo.{},patch_points=np.array(patch_points),b_coef=np.array(b_sol),sol_mat=np.array(solution_matrix))))".format(fn))
                exec("DD.append(partial(foo.{},patch_points=np.array(patch_points),b_coef=np.array(b_sol),sol_mat=np.array(solution_matrix)))".format(fn))
        self.pickled_DD = pickled_DD
        self.DD = DD
        #print('get properties')
        self._get_properties_(resolution=resolution,k=k,level=level,visualize=visualize,buffer=buffer)
        """
        DD = []
        CDD = []
        for f in functions:
            if len(self.patches) > 1:
                DD.append(partial(f,KDTree=KDTree,patch_points=np.array(patch_points),
                                  b_coef=np.array(b_sol),h=h,q=q,sol_mat=np.array(solution_matrix),
                                  patch_1=np.array(patch_x),patch_2=np.array(patch_y),
                                  patch_3=np.array(patch_z)))
            else:
                tmp_DD = partial(foo,patch_points=np.array(patch_points),b_coef=np.array(b_sol),
                                 sol_mat=np.array(solution_matrix))
                CDD.append(dumps(tmp_DD))
                DD.append(tmp_DD)
        if len(self.patches) > 1:
            func = partial(function,KDTree=KDTree,patch_points=np.array(patch_points),
                           b_coef=np.array(b_coef),h=h,q=q,sol_mat=np.array(solution_matrix),
                           patch_x=np.array(patch_x),patch_y=np.array(patch_y),
                           patch_z=np.array(patch_z))
            func_marching = partial(function_marching,patch_points=patch_points_m,
                                    a_coef=a_coef,b_coef=b_coef,h=h,q=q,
                                    patch_x=np.array(patch_x),patch_y=np.array(patch_y),
                                    patch_z=np.array(patch_z))
            grad = partial(gradient,KDTree=KDTree,patch_points=np.array(patch_points),
                           b_coef=np.array(b_coef),h=h,q=q,sol_mat=np.array(solution_matrix),
                           patch_x=np.array(patch_x),patch_y=np.array(patch_y),
                           patch_z=np.array(patch_z))
            self.function = func
            self.function_marching = func_marching
            self.gradient = grad
        self.DD = DD
        self.CDD = CDD
        """
    def build_individual(self,idx,h=0,q=1):
        func_marching = partial(function_marching,patch_points=[self.patch_points_m[idx]],
                                a_coef=[self.a_coef[idx]],b_coef=[self.b_coef[idx]],h=h,q=q,
                                patch_x=np.array([self.patch_x[idx]]),patch_y=np.array([self.patch_y[idx]]),
                                patch_z=np.array([self.patch_z[idx]]))
        return func_marching

    def within(self,x,y,z,k,level=0):
        return self.DD[0]([x,y,z,k]) < level

    def within_mesh(self,x,y,z,k):
        point = pv.PolyData(np.array([x,y,z]))
        select = point.select_enclosed_points(self.pv_polydata)
        return bool(select['SelectedPoints'][0])

    def pick_in_cell(self,cell_idx):
        simplex = self.tet_pts[self.tet_verts[cell_idx,:],:]
        point = np.array(generate(simplex,1))
        return point,None

    def pick(self,**kwargs):
        homogeneous = kwargs.get('homogeneous',False)
        replacement = kwargs.get('replacement',True)
        size = kwargs.get('size',1)
        vert_applied = kwargs.get('vert',False)
        verts = kwargs.get('verts',[])
        if homogeneous:
            point = None
            if not vert_applied:
                rdx = np.random.Generator(np.random.PCG64()).choice(list(range(self.tet_verts.shape[0])),size,p=self.norm_ele_vol,replace=replacement)
            else:
                rdx = verts
            simplex = self.tet_pts[self.tet_verts[rdx,:],:]
            point = np.array(generate(simplex,size))
            return point,None
        id = np.random.choice(self.pv_polydata_surf.n_faces)
        faces = self.pv_polydata_surf.faces.reshape(-1,self.pv_polydata_surf.faces[0]+1)
        pts = self.pv_polydata_surf.points[faces[id,1:],:]
        a1 = np.random.random(1)
        a2 = np.random.random(1)
        if (a1+a2) > 1:
            a1 = 1 - a1
            a2 = 1 - a2
        A = 1 - a1 - a2
        B = a1
        C = a2
        pts = pts*np.array([A,B,C])
        pts = np.sum(pts,axis=0)
        pts = np.array(pts)
        #x = np.random.uniform(self.x_range[0],self.x_range[1])
        #y = np.random.uniform(self.y_range[0],self.y_range[1])
        #z = np.random.uniform(self.z_range[0],self.z_range[1])
        #k     = kwargs.get('k',len(self.patches))
        #layer = kwargs.get('layer',0)
        #x0    = kwargs.get('x0',np.array([x,y,z]))
        #f     = lambda x: (self.DD[0]([x[0],x[1],x[2],k])-layer)**2
        #df    = lambda x: 2*(self.DD[0]([x[0],x[1],x[2],k])-layer)*self.DD[1]([x[0],x[1],x[2],k])
        #ddf   = lambda x,y,z: 2*np.dot(self.DD[1]([x,y,z,k],self.DD[1]([x,y,z,k]))+2*self.DD[2]([x,y,z,k])*(self.DD[0]([x,y,z,k])-layer)
        #res   = optimize.minimize(f,x0,method='L-BFGS-B')
        return pts,None

    def path(self,start,distance,steps,dive=0.01,theta_steps=40):
        start_layer = self.DD[0]([start[0],start[1],start[2],len(self.patches)])
        theta = np.linspace(0,2*np.pi,theta_steps)
        path = [start]
        path_d = [0]
        path_layer = [start_layer]
        point = start
        while np.sum(path_d) < distance:
            normal = self.DD[1]([point[0],point[1],point[2],max(round(len(self.patches)),2)])
            t1,t2,unit_normal = tangent_basis(normal,point)
            next_steps = path[-1] + (distance/steps)*(t1*np.cos(theta).reshape(-1,1)+t2*np.sin(theta).reshape(-1,1)) #-(dive/steps)*(unit_normal)
            distances  = np.linalg.norm(next_steps - path[-1],axis=1)
            distances1 = np.linalg.norm(next_steps - path[0],axis=1)+np.linalg.norm(next_steps - path[-1],axis=1)
            next_step  = next_steps[np.argmax(distances1),:]
            next_step,next_layer = self.pick(layer=path_layer[-1]-(dive/steps),x0=next_step)
            path_layer.append(next_layer)
            path_d.append(distances[np.argmax(distances)])
            path.append(next_step)
            point = next_step
        return path,path_d,path_layer

    def mesh(self,path,distance,steps,dive=0.01,theta_steps=40,others=None):
        center = round(len(path)*0.5)
        start  = path[center]
        start_layer = self.DD[0]([start[0],start[1],start[2],len(self.patches)])
        ###################
        ###################
        theta = np.linspace(0,2*np.pi,theta_steps)
        center_path = [start]
        path_d = []
        path_layer = [start_layer]
        point = start
        direction = (path[center+1] - path[center-1])/np.linalg.norm(path[center+1] - path[center-1])
        while np.sum(path_d) < distance:
            normal = self.DD[1]([point[0],point[1],point[2],max(round(len(self.patches)*0.1),2)])
            t1,t2,unit_normal = tangent_basis(normal,point)
            next_steps = center_path[-1] + (distance/steps)*(t1*np.cos(theta).reshape(-1,1)+
                                                             t2*np.sin(theta).reshape(-1,1)) #-(dive/steps)*(unit_normal)
            distances  = np.linalg.norm(next_steps - center_path[-1],axis=1)
            if len(center_path) == 1:
                distances1 = np.linalg.norm(next_steps - center_path[0],axis=1)+np.linalg.norm(next_steps - path[center+1],axis=1)+np.linalg.norm(next_steps - path[center-1],axis=1)
            else:
                #distances1 = np.linalg.norm(next_steps - center_path[0],axis=1)+np.linalg.norm(next_steps - center_path[-1],axis=1)+\
                if others is None:
                    distances1 = (1 - np.sum(path_d)/distance)*np.linalg.norm(next_steps - center_path[0],axis=1)+np.sum(path_d)/distance*np.linalg.norm(next_steps - center_path[-1],axis=1)
                else:
                    other_dist,_ = others.query(next_steps,k=1)
                    distances1 = (1 - np.sum(path_d)/distance)*np.linalg.norm(next_steps - center_path[0],axis=1)+np.sum(path_d)/distance*np.linalg.norm(next_steps - center_path[-1],axis=1)+\
                                 (other_dist)
            next_step  = next_steps[np.argmax(distances1),:]
            next_step,next_layer = self.pick(layer=path_layer[-1]-(dive/steps),x0=next_step)
            path_layer.append(next_layer)
            path_d.append(distances[np.argmax(distances)])
            center_path.append(next_step)
            point = next_step
        past_up = []
        past_down = []
        mesh = []
        #mesh.extend(center_path)
        #mesh.extend(path)
        up_past = []
        down_past = []
        past_point_up = []
        past_point_down = []
        past_line = center_path
        #Upper triangle Mesh
        #print('Max Center Path: {}'.format(len(path[center+1:])))
        for i in range(len(path[center+1:-1])):
            ii = i + center +1
            point = path[ii]
            line = [point]
            percent = round(min((1-i/len(path[center+1:])),1)*len(center_path))
            #print(percent)
            for j in range(1,percent):
                normal = self.DD[1]([point[0],point[1],point[2],max(round(len(self.patches)*0.1),2)])
                m1 = (past_line[j] - past_line[j-1])/np.linalg.norm(past_line[j] - past_line[j-1])
                m2 = (point - past_line[j-1])/np.linalg.norm(line[-1] - past_line[j-1])
                predicted = past_line[j-1] + m1*(distance/steps)+m2*(distance/steps)
                t1,t2,unit_normal = tangent_basis(normal,point)
                next_steps = point + (distance/steps)*(t1*np.cos(theta).reshape(-1,1)+t2*np.sin(theta).reshape(-1,1))
                distances = np.linalg.norm(next_steps - predicted,axis=1)
                line.append(next_steps[np.argmin(distances),:])
                point = line[-1]
            past_line = line
            mesh.extend(line)
            #show_mesh(np.array(mesh),np.array(path),np.array(center_path))
        past_line = center_path
        #Lower Triangle Mesh
        for i in range(len(path[1:center-1])):
            ii = center -1-i
            point = path[ii]
            line = [point]
            percent = round(min((1-i/len(path[:center-1])),1)*len(center_path))
            for j in range(1,percent):
                normal = self.DD[1]([point[0],point[1],point[2],max(round(len(self.patches)*0.1),2)])
                m1 = (past_line[j] - past_line[j-1])/np.linalg.norm(past_line[j] - past_line[j-1])
                m2 = (point - past_line[j-1])/np.linalg.norm(line[-1] - past_line[j-1])
                predicted = past_line[j-1] + m1*(distance/steps)+m2*(distance/steps)
                t1,t2,unit_normal = tangent_basis(normal,point)
                next_steps = point + (distance/steps)*(t1*np.cos(theta).reshape(-1,1)+t2*np.sin(theta).reshape(-1,1))
                distances = np.linalg.norm(next_steps - predicted,axis=1)
                line.append(next_steps[np.argmin(distances),:])
                point = line[-1]
            past_line = line
            mesh.extend(line)
        #proximal = path[0]
        #distal = path[-1]
        #terminal = center_path[-1]
        #mesh.insert(0,terminal)
        #mesh.insert(0,distal)
        #mesh.insert(0,proximal)
        mesh = np.array(mesh)
        mesh_tree = cKDTree(mesh)
        graph = mesh_tree.sparse_distance_matrix(mesh_tree,((distance/steps)**2+(dive/steps)**2)*(1/2))
        return mesh,path,center_path,path_d,graph

    def _get_properties_(self,**kwargs):
        resolution = kwargs.get('resolution',20)
        k          = kwargs.get('k',2)
        level      = kwargs.get('level',0)
        visualize  = kwargs.get('visualize',False)
        buffer     = kwargs.get('buffer',1)
        self.polydata,self.tet_pts,self.tet_verts,self.ele_vol,self.tet = marching_cubes_pv(self,resolution=resolution,k=k,level=level,visualize=visualize,buffer=buffer)
        self.cell_lookup = cKDTree(self.tet.grid.cell_centers().points)
        self.pv_polydata = pv.PolyData(var_inp=self.polydata)
        self.pv_polydata_surf = self.pv_polydata.extract_surface()
        nodes_check = set([])
        nodes = []
        idx = [[0,1],[1,2],[2,3],[3,1]]
        lengths = []
        for tet in self.tet_verts:
            for dx in idx:
                if tuple([tet[dx[0]],tet[dx[1]]]) in nodes_check:
                    continue
                nodes_check.add(tuple([tet[dx[0]],tet[dx[1]]]))
                nodes_check.add(tuple([tet[dx[1]],tet[dx[0]]]))
                nodes.append([tet[dx[0]],tet[dx[1]]])
                nodes.append([tet[dx[1]],tet[dx[0]]])
                l = np.linalg.norm(self.tet_pts[tet[dx[0]]] - self.tet_pts[tet[dx[1]]])
                lengths.append(l)
                lengths.append(l)
        M = np.array(nodes)
        graph = csr_matrix((lengths,(M[:,0],M[:,1])),shape=(max(M[:,0])+1,max(M[:,1])+1))
        self.graph = graph
        self.shortest_algorithm = lambda ind: shortest_path(csgraph=graph,directed=False,indices=ind,return_predecessors=True)
        def get_path(i,j):
            ds, Pr = shortest_path(csgraph=graph,directed=False,indices=i,return_predecessors=True)
            path = [j]
            lengths = []
            k = j
            while Pr[k] != -9999:
                path.append(Pr[k])
                lengths.append(ds[k])
                k = Pr[k]
            #lengths.append(ds[-9999])
            path = path[::-1]
            lengths = lengths[::-1]
            lines = []
            for jdx in range(len(path)-1):
                lines.append([path[jdx],path[jdx+1]])
            return path,lengths,lines
        def get_cells(path,grid=self.tet.grid):
            cells = []
            cell_sets = []
            for idx,i in enumerate(path):
                cells.append(grid.extract_points(i))
                cell_sets.append(set(cells[-1].cell_data['vtkOriginalCellIds']))
                if idx > 0:
                    cell_sets[-2] = cell_sets[-2].difference(cell_sets[-1])
                    cell_sets[-1] = cell_sets[-1].difference(cell_sets[-2])
            cell_sets = [list(cs) for cs in cell_sets]
            return cell_sets
        def get_cell_points(cells,verts=self.tet_verts,pts=self.tet_pts):
            return pts[verts[cells]]
        def point_function(s1,s2,s3,n,cell_pts):
            dn = 1/cell_pts.shape[0]
            N  = int(n // dn)
            if N > cell_pts.shape[0]-1:
                N =  cell_pts.shape[0]-1
            if N < 0:
                N = 0
            if s1 + s2 > 1:
                s1 = 1.0 - s1
                s2 = 1.0 - s2
            if s2 + s3 > 1:
                tmp = s3
                s3 = 1.0 - s1 - s2
                s2 = 1.0 - tmp
            elif s1 + s2 + s3 > 1:
                tmp = s3
                s3 = s1 + s2 + s3 - 1.0
                s1 = 1.0 - s2 - tmp
            s4 = 1 - s1 - s2 - s3
            #print(s4)
            #print(s1.VALUE)
            #print(s2.VALUE)
            #print(s3.VALUE)
            #print(s4.value)
            S  = np.array([s1,s2,s3,s4]).reshape(-1,1)
            #id = 0
            #for i in range(cell_pts.shape[0]):
            #    if n.value == i:
            #        id = i
            #        break
            #P1  = s1*cell_pts[id,0,0] + s2*cell_pts[id,1,0] + s3*cell_pts[id,2,0] + s4*cell_pts[id,3,0]
            #P2  = s1*cell_pts[id,0,1] + s2*cell_pts[id,1,1] + s3*cell_pts[id,2,1] + s4*cell_pts[id,3,1]
            #P3  = s1*cell_pts[id,0,2] + s2*cell_pts[id,1,2] + s3*cell_pts[id,2,2] + s4*cell_pts[id,3,2]
            #print(id)
            P  = cell_pts[N]*S
            return np.sum(P,axis=0)

        def get_point_functions(cell_sets):
            point_functions = []
            for i in range(len(cell_sets)):
                cell_pts = get_cell_points(cell_sets[i])
                f = partial(point_function,cell_pts=cell_pts)
                point_functions.append(f)
            return point_functions
        def path_function(x,start,end,point_functions):
            a = [start]
            b = []
            #l = []
            for i in range(len(point_functions)):
                x_tmp = x[i*4:i*4+4]
                pt = point_functions[i](x_tmp[0],x_tmp[1],x_tmp[2],x_tmp[3])
                #l.append(((pt[0] - a[-1][0])**2+(pt[1] - a[-1][1])**2+(pt[2] - a[-1][2])**2)**(1/2))
                a.append(pt)
                b.append(pt)
            b.append(end)
            #l.append(((pt[0] - end[0])**2+(pt[1] - end[1])**2+(pt[2] - end[2])**2)**(1/2))
            A = np.array(a)
            B = np.array(b)
            L = np.sum(np.linalg.norm(B-A,axis=1))
            #L = sum(l)
            return L
        def build_path_function(i,j,start,end):
            path,_,_ = get_path(i,j)
            cell_sets = get_cells(path)
            point_functions = get_point_functions(cell_sets)
            pf = partial(path_function,start=start,end=end,point_functions=point_functions)
            return pf,point_functions
        def optimize_path(pf,func_num,niter=200,disp=True):
            x0 = np.zeros(func_num*4)
            bounds = []
            for i in range(len(x0)):
                bounds.append(tuple([0,0.99]))
            kwargs = {'method':"L-BFGS-B",
                      'bounds':bounds}
            #m = GEKKO(remote=False)
            #x = []
            #for i in range(len(funcs)):
            #    x.append(m.Var(value=0,lb=0,ub=1))
            #    x.append(m.Var(value=0,lb=0,ub=1))
            #    x.append(m.Var(value=0,lb=0,ub=1))
            #    x.append(m.Var(value=0,lb=0,ub=len(funcs[i].keywords['cell_pts']),integer=True))
            #x = m.Array(m.Var,len(x0),lb=0,ub=1,integer=True)
            #m.Minimize(pf(x))
            #m.options.SOLVER=1
            #m.solver_options = ['minlp_maximum_iterations 500', \
            #        # minlp iterations with integer solution
            #        'minlp_max_iter_with_int_sol 10', \
            #        # treat minlp as nlp
            #        'minlp_as_nlp 0', \
            #        # nlp sub-problem max iterations
            #        'nlp_maximum_iterations 50', \
            #        # 1 = depth first, 2 = breadth first
            #        'minlp_branch_method 1', \
            #        # maximum deviation from whole number
            #        'minlp_integer_tol 0.05', \
            #        # covergence tolerance
            #        'minlp_gap_tol 0.01']
            #m.solve()

            res = optimize.basinhopping(pf,x0,niter=niter,minimizer_kwargs=kwargs,disp=disp)
            return res
        def optimized_points(res,point_functions,start,end):
            pts = [start]
            a = [start]
            b = []
            sol = res.x
            for i in range(len(point_functions)):
                x_tmp = sol[i*4:i*4+4]
                pts.append(point_functions[i](x_tmp[0],x_tmp[1],x_tmp[2],x_tmp[3]))
                a.append(pts[-1])
                b.append(pts[-1])
            pts.append(end)
            b.append(end)
            A = np.array(a)
            B = np.array(b)
            L = np.linalg.norm(B-A,axis=1)
            return pts,L
        def find_best_path(start,end,grid=self.tet.grid,niter=200,disp=True):
            i = grid.find_closest_point(start)
            j = grid.find_closest_point(end)
            pf,funcs = build_path_function(i,j,start,end)
            res = optimize_path(pf,len(funcs),niter=niter,disp=disp)
            pts,L = optimized_points(res,funcs,start,end)
            ## check for unique
            _,u_pts_idx = np.unique(pts,axis=0,return_index=True)
            u_pts_idx = np.flip(u_pts_idx)
            rm_idx = np.argwhere(np.diff(u_pts_idx)>1)
            return pts,L,res,pf,funcs,rm_idx
        def spline_resample(path_pts):
            if len(path_pts) == 2:
                path = splprep(path_pts.T,k=1,s=0)
            elif len(path_pts) == 3:
                path = splprep(path_pts.T,k=2,s=0)
            else:
                path = splprep(path_pts.T,s=0)
            return path
        def resample(path,niter=20,disp=True):
            def resample_optimizer(on_off,path_pts=path,within=self.pickled_DD[0]):
                on_off = on_off > 0.5
                last_pts = path_pts[0]
                end_pts = path_pts[-1]
                cost = 0
                within = loads(within)
                cost += np.sum(on_off)
                for idx,i in enumerate(on_off):
                    if i == 1:
                        subdivision = 4
                        for j in range(1,2*subdivision):
                            tmp = (last_pts + path_pts[idx+1])*(j/(2*subdivision))
                            if within((tmp[0],tmp[1],tmp[2],2)) > 0.15:
                                cost += 100
                        last_pts = path_pts[idx+1]
                subdivision  = 4
                for j in range(1,2*subdivision):
                    tmp = (last_pts + path_pts[-1])*(j/(2*subdivision))
                    if within((tmp[0],tmp[1],tmp[2],2)) > 0.15:
                        cost += 100
                return cost
            x0 = np.ones(len(path)-2)
            bounds = []
            for i in range(len(x0)):
                bounds.append(tuple([0,1]))
            kwargs = {'method':"L-BFGS-B",
                      'bounds':bounds}
            #print(len(path))
            #m = GEKKO(remote=False)
            #x = m.Array(m.Var,len(x0),lb=0,ub=1,integer=True)
            #m.Minimize(resample_optimizer(x))
            #m.options.SOLVER=1
            #m.solve()
            #on_off = x
            #print(on_off)
            res = optimize.basinhopping(resample_optimizer,x0,niter=100,minimizer_kwargs=kwargs,disp=disp)
            on_off = res.x > 0.5
            #print(res[1])
            new_path = [path[0]]
            new_lengths = []
            for idx,i in enumerate(on_off):
                if i:
                    new_path.append(path[idx+1])
                    new_lengths.append(np.linalg.norm(new_path[-1]-new_path[-2]))
            new_path.append(path[-1])
            new_lengths.append(np.linalg.norm(new_path[-1]-new_path[-2]))
            #print(len(new_path))
            return np.array(new_path), new_lengths
        self.get_shortest_path = get_path
        self.build_path_function = build_path_function
        self.optimize_path = optimize_path
        self.optimized_points = optimized_points
        self.find_best_path = find_best_path
        self.resample = resample
        self.total_ele_vol = sum(self.ele_vol)
        self.norm_ele_vol = self.ele_vol/self.total_ele_vol
        #volume,surface_area = properties(self.polydata)
        volume = self.pv_polydata.volume
        surface_area = self.pv_polydata.area
        self.volume= volume
        self.surface_area = surface_area

    def subtract(self,other_surface):
        self.subtracted_volumes.append(other_surface)
        def d_0(xyz,F=self.DD[0],S=other_surface.DD[0]):
            return F(xyz)-(S(xyz)<0)*(F(xyz)+S(xyz))
        self.DD[0] = d_0
        def d_1(xyz,F=self.DD[0],S=other_surface.DD[0],dF=self.DD[1],dS=other_surface.DD[1]):
            return dF(xyz)-(S(xyz)<0)*(1-S(xyz))*(dF(xyz)+dS(xyz))
        self.DD[1] = d_1
        def d_2(xyz,F=self.DD[0],S=other_surface.DD[0],ddF=self.DD[2],ddS=self.DD[2]):
            return ddF(xyz)-(S(xyz)<0)*(1-S(xyz))*(ddF(xyz)+ddS(xyz))
        self.DD[2] = d_2
        self.volume = self.volume - other_surface.volume
        self.surface_area = self.surface_area - other_surface.surface_area

    def add(self,other_surface):
        self.added_volumes.append(other_surface)
        def d_0(xyz,F=self.DD[0],S=other_surface.DD[0]):
            return F(xyz)+(S(xyz)<0)(F(xyz)>0)*(S(xyz))
        self.DD[0] = d_0
        def d_1(xyz,F=self.DD[0],S=other_surface.DD[0],dF=self.DD[1],dS=other_surface.DD[1]):
            return dF(xyz)+(S(xyz)<0)*(F(xyz)>0)*dS(xyz)
        self.DD[1] = d_1
        def d_2(xyz,F=self.DD[0],S=other_surface.DD[0],ddF=self.DD[2],ddS=self.DD[2]):
            return ddF(xyz)+(S(xyz)<0)*(F(xyz)>0)*ddS(xyz)
        self.DD[2] = d_2

    def export_polydata(self,file):
        writer = vtk.vtkXMLPolyDataWriter()
        writer.SetFileName(file)
        writer.SetInputDataObject(self.polydata)
        writer.Write()

def marching_cubes(surface_object,resolution=20,k=2,level=0,visualize=False,buffer=1):
    """
    Takes an interpolated volume and performs
    a descritization on a selected hyperplane.
    """
    surface_algorithm = vtk.vtkMarchingCubes()
    image_data        = vtk.vtkImageData()
    colors = vtk.vtkNamedColors()
    X,Y,Z = np.mgrid[surface_object.x_range[0]-buffer:surface_object.x_range[1]+buffer:resolution*1j,
                     surface_object.y_range[0]-buffer:surface_object.y_range[1]+buffer:resolution*1j,
                     surface_object.z_range[0]-buffer:surface_object.z_range[1]+buffer:resolution*1j]
    #print(X.shape[1],X.shape[0],X.shape[2])
    SHAPE = (resolution,resolution,resolution)
    Xs = ((-(surface_object.x_range[0]-buffer)+(surface_object.x_range[1]+buffer))/(resolution-1))
    Ys = ((-(surface_object.y_range[0]-buffer)+(surface_object.y_range[1]+buffer))/(resolution-1))
    Zs = ((-(surface_object.z_range[0]-buffer)+(surface_object.z_range[1]+buffer))/(resolution-1))
    SPACING = (Zs,Ys,Xs)
    Xf = X.flatten()
    Yf = Y.flatten()
    Zf = Z.flatten()
    Kf = np.ones(Xf.shape,dtype=int)*k
    ORIGIN = (min(Zf),min(Yf),min(Xf))
    #Sample Implicit Function
    #print('Sampling Implicit Volume...')
    Vf = []
    for i in zip(Zf,Yf,Xf,Kf):
        Vf.append(surface_object.DD[0](i))
    Vf = np.array(Vf)
    surface_object.depth_interior_min = min(Vf)
    Vf = Vf.reshape(X.shape)
    #grid = pv.UniformGrid(dims=SHAPE,spacing=SPACING,origin=ORIGIN)
    #out = grid.contour(1,scalars=Vf,rng=[0,0],method='marching_cubes')
    #out.plot(color='tan',smooth_shading=True)
    #print('Converting to VTK Image Data...')
    #Vf = np.array(Vf)
    #Vf = Vf.reshape(X.shape)
    pointsdata = vtk.vtkPoints()
    for i in range(surface_object.points.shape[0]):
        id = pointsdata.InsertNextPoint(surface_object.points[i,0],surface_object.points[i,1],surface_object.points[i,2])
    pt_poly = vtk.vtkPolyData()
    pt_poly.SetPoints(pointsdata)
    vertexfilter = vtk.vtkVertexGlyphFilter()
    vertexfilter.SetInputData(pt_poly)
    vertexfilter.Update()
    PTS = vtk.vtkPolyData()
    PTS.ShallowCopy(vertexfilter.GetOutput())
    pt_mapper = vtk.vtkPolyDataMapper()
    pt_mapper.SetInputData(PTS)
    pt_actor = vtk.vtkActor()
    pt_actor.SetMapper(pt_mapper)
    pt_actor.GetProperty().SetPointSize(10)
    pt_actor.GetProperty().SetColor(colors.GetColor3d('Red'))
    #Vf = np.flipud(Vf)
    #Vf = np.flip(Vf,1)
    data_vtk = numpy_support.numpy_to_vtk(Vf.ravel(),deep=1,array_type=vtk.VTK_FLOAT)
    image_data.SetDimensions(SHAPE)
    image_data.SetSpacing(SPACING)
    image_data.SetOrigin(ORIGIN)
    #image_data.SetOrigin(-np.max(Xf)/2,np.max(Yf)/2,-np.max(Zf)/2)
    image_data.GetPointData().SetScalars(data_vtk)
    image_data.Modified()
    #print('Marching...')
    surface_algorithm.SetInputData(image_data)
    surface_algorithm.ComputeNormalsOn()
    surface_algorithm.SetValue(0,level)

    if visualize:
        colors = vtk.vtkNamedColors()

        renderer = vtk.vtkRenderer()
        renderer.SetBackground(colors.GetColor3d('White'))

        render_window = vtk.vtkRenderWindow()
        render_window.AddRenderer(renderer)
        render_window.SetWindowName('Marching Cubes')

        interactor = vtk.vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(surface_algorithm.GetOutputPort())
        mapper.ScalarVisibilityOff()

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)

        actor.GetProperty().SetColor(colors.GetColor3d('MistyRose'))

        renderer.AddActor(actor)
        renderer.AddActor(pt_actor)
        render_window.Render()
        interactor.Start()
    else:
        pass
        #surface_algorithm.Update()
    if surface_object.polydata is not None:
        volume_polydata = surface_object.polydata
    else:
        surface_algorithm.Update()
        volume_polydata = surface_algorithm.GetOutput()
    transform = vtk.vtkTransform()
    pv_tmp = pv.PolyData(var_inp=volume_polydata)
    if not pv_tmp.is_all_triangles:
        #print('Triangulating mesh...')
        pv_tmp.triangulate(inplace=True)
        pv_tmp.plot(show_edges=True,opacity=0.5)
    tet = tetgen.TetGen(pv_tmp)
    try:
        nodes, verts = tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
    except:
        tet.make_manifold()
        nodes, verts = tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
    cells = tet.grid.compute_cell_sizes()
    ele_vol = cells.cell_data["Volume"]
    return volume_polydata, nodes, verts, ele_vol, tet


def marching_cubes_pv(surface_object,resolution=20,k=2,level=0,visualize=False,
                      buffer=1,method='marching_cubes',compute_normals=False,
                      compute_gradients=False,verbose=False,tetgen_verbose=0):
    x_min = surface_object.x_range[0] - buffer
    y_min = surface_object.y_range[0] - buffer
    z_min = surface_object.z_range[0] - buffer
    x_max = surface_object.x_range[1] + buffer
    y_max = surface_object.y_range[1] + buffer
    z_max = surface_object.z_range[1] + buffer
    dims    = (resolution,resolution,resolution)
    spacing = (abs(x_max-x_min)/resolution,abs(y_max-y_min)/resolution,abs(z_max-z_min)/resolution)
    origin  = (x_min,y_min,z_min)
    grid    = pv.UniformGrid(dims=dims,spacing=spacing,origin=origin)
    x,y,z   = grid.points.T
    values  = []
    if verbose:
        bar = trange
    else:
        bar = range
    for i in bar(len(x)):
        values.append(surface_object.DD[0]([x[i],y[i],z[i],k]))
    values = np.array(values)
    surface_object.depth_interior_min = min(values)
    mesh = grid.contour([level],values,method=method,
                        compute_gradients=compute_gradients,
                        compute_normals=compute_normals)
    mesh.set_active_scalars(mesh.array_names[0])
    tet = tetgen.TetGen(mesh)
    tet.make_manifold(verbose=verbose)
    nodes, verts = tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5,verbose=tetgen_verbose)
    cells = tet.grid.compute_cell_sizes()
    ele_vol = cells.cell_data["Volume"]
    surf   = mesh.extract_surface()
    mf = pymeshfix.MeshFix(surf)
    mf.repair(verbose=verbose)
    surf = mf.mesh
    surf.save(os.getcwd()+os.sep+'temp.vtp')
    reader = vtk.vtkXMLPolyDataReader()
    reader.SetFileName(os.getcwd()+os.sep+'temp.vtp')
    reader.Update()
    surf = reader.GetOutput()
    os.remove(os.getcwd()+os.sep+'temp.vtp')
    return surf, nodes, verts, ele_vol, tet

def properties(polydata_object):
    MASS = vtk.vtkMassProperties()
    MASS.SetInputData(polydata_object)
    MASS.Update()
    return MASS.GetVolume(),MASS.GetSurfaceArea()

@nb.jit(nopython=True)
def generate(points,num):
    #r1,r2,r3 = np.random.random((3,num))
    s = np.zeros(num*4)
    for i in range(num):
        r1,r2,r3 = np.random.random(3)
        if r1 + r2 > 1:
            r1 = 1.0 - r1
            r2 = 1.0 - r2
        if r2 + r3 > 1:
            tmp = r3
            r3 = 1.0 - r1 - r2
            r2 = 1.0 - tmp
        elif r1 + r2 + r3 > 1:
            tmp = r3
            r3 = r1 + r2 + r3 - 1.0
            r1 = 1.0 - r2 - tmp
        a = 1.0 - r1 - r2 - r3
        idx_1 = i*4
        idx_2 = i*4+1
        idx_3 = i*4+2
        idx_4 = i*4+3
        s[idx_1] = a
        s[idx_2] = r1
        s[idx_3] = r2
        s[idx_4] = r3
    s = s.reshape((-1,1))
    points = s*points.reshape(num*4,3)
    points = np.sum(points.reshape(num,4,3),axis=1)
    return points
"""
@nb.jit(nopython=True)
def lines_to_cells(p0,centers,subdivisions):
    points = []
    for i in p0:
        for j in centers:
            for k in range(1,2*subdivisions):
                points.append((i+j)*(k/(2*subdivisions)))
"""



import numpy as np
from .basis import *
from .calculate_length import *
from .update import *
from .calculate_radii import *


def set_root_homogeneous_nonconvex(data, boundary, Qterm, gamma, nu,
                                   Pperm, Pterm, start,direction):
    pass



geometry="""num_samples = {}
use_distance_alignment_method = {}
for contour_polydata_list in contour_polydata:
    for idx in range(len(contour_polydata_list)):
        contour_polydata_list[idx] = geometry.interpolate_closed_curve(polydata=contour_polydata_list[idx],number_of_points=num_samples)
        if idx != 0:
            contour_polydata_list[idx] = geometry.align_profile(contour_polydata_list[idx-1],contour_polydata_list[idx],use_distance_alignment_method)
"""



import numpy as np

"""
Function to approximate wall
shear stress for a cylindrical
pipe under poiseuille assumptions
with ideal circular cross-sections.

Multifidelity Estimators For Coronary Circulation
Models under clinically-informed Data Uncertainty

Reference: https://arxiv.org/pdf/1911.11266.pdf
"""

def wss(*args,**kwargs):
    """
    Args
    ----

    Kwargs
    ------

    Returns
    -------
    """
    object = args[0]
    radius = object.data[:,21]
    flow   = object.data[:,22]
    dynamic_viscosity = object.parameters['nu']
    wss = (4*dynamic_viscosity*flow)/(np.pi*radius**3)
    return wss



import numba as nb
import numpy as np
from .add_edge import *
from .basis import *
from .calculate_length import *
from copy import deepcopy
import time

#@nb.jit(nopython=True,cache=True,nogil=True)
def add_bifurcation(tree,parent_edge,candidate_point,results,optimum,isforest=False):
    ########### UNPACK LOCAL OPTIMIZATION DATA ##############
    start = time.time()
    R0  = results[0]
    L0 = results[1]
    f_terminal = results[2]
    f_sister   = results[3]
    bif_points   = results[5]
    RR = np.array(results[6])
    RL    = np.array(results[7])
    bifs               = np.array(results[8])
    flows              = np.array(results[9])
    main_idx           = np.array(results[10])
    alt_idx            = np.array(results[11])
    main_scale         = np.array(results[12])
    alt_scale          = np.array(results[13])
    R_terminal = results[14]
    R_sister   = results[15]
    if isforest:
        sub_division_map    = deepcopy(tree.sub_division_map)
        sub_division_index  = deepcopy(tree.sub_division_index)
    else:
        sub_division_map    = tree.sub_division_map
        sub_division_index  = tree.sub_division_index
    Qterm               = tree.parameters['Qterm']
    tree.time['add_1'].append(time.time()-start)
    ###########################################################
    ############### Copy Data and Make New ####################
    start = time.time()
    if isforest:
        data = deepcopy(tree.data)
    else:
        data = tree.data
    #if sum(tree.data[tree.data[:,-1]==-1]) == 0:
    #    nosegments = True
    #else:
    #    nosegments = False
    #    segment_data = deepcopy(tree.data[tree.data[:,-1]==-1])
    nosegments = True
    data = np.vstack((data,np.zeros((2,data.shape[1])))) ## would need to be changed for adding non-linear vessels
    bifurcation_point = bif_points[optimum]
    add_edge(data[:-1,:],bifurcation_point,candidate_point,parent_edge,Qterm)
    data[-2,17] = np.float(parent_edge)
    terminal_sub_division = [-1]
    downstream_node = data[parent_edge,3:6]
    downstream_flow = data[parent_edge,22].item()
    bifurcation_node = data[-2,18].item()
    distal_node = data[parent_edge,19].item()
    left = data[parent_edge,15].item()
    right = data[parent_edge,16].item()
    lbif = data[parent_edge,23]
    rbif = data[parent_edge,24]
    downstream_rr = data[parent_edge,27]
    add_edge(data,bifurcation_point,downstream_node,parent_edge,
             downstream_flow,proximal_idx=bifurcation_node,
             distal_idx=distal_node,left_child=left,
             right_child=right)
    data[-1,17] = np.float(parent_edge)
    sister_sub_division = [-1]
    tree.time['add_2'].append(time.time()-start)
    #pidx_start = sub_division_index[parent_edge]
    #pidx_end = sub_division_index[parent_edge+1]
    start_time = time.time()
    if left > 0 or right > 0: # if this vessel is not a terminal it will have a subdivision map
        start = sub_division_index[parent_edge] #start of the
        if parent_edge+1 == len(sub_division_index):
            end = None
        else:
            end = sub_division_index[parent_edge+1]
        k = sub_division_map[start+1:end]
        data[k,28] = (data[k,28]/data[parent_edge,28])*f_sister[optimum] # replace the old bifurcation with the new scaling
        data[k,26] += 1 # incrament the depth of all subdivision vessels
        sister_sub_division.extend(k)
        data[-1,23] = lbif
        data[-1,24] = rbif
    data[-1,28] = f_sister[optimum]
    data[-1,27] = downstream_rr
    data[-2,28] = f_terminal[optimum]
    data[-1,25] = R_sister[optimum]
    data[-2,25] = R_terminal[optimum]
    tree.time['add_3'].append(time.time()-start_time)
    if int(left) > 0:
        data[int(left),17] = data[-1,-1]
    if int(right) > 0:
        data[int(right),17] = data[-1,-1]
    data[parent_edge,3:6] = bifurcation_point
    data[parent_edge,19] = bifurcation_node
    data[parent_edge,15] = data[-1,-1]
    data[parent_edge,16] = data[-2,-1]
    new_terminal_edge = int(data[-2,-1])
    new_sister_edge = int(data[-1,-1])
    basis(data,parent_edge.item())
    length(data,parent_edge.item())
    sub_division_map.extend(terminal_sub_division) #should be [-1]
    sub_division_index = np.append(sub_division_index,len(sub_division_map)-1)
    sub_division_index = np.append(sub_division_index,len(sub_division_map))
    sub_division_map.extend(sister_sub_division) #should be a copy of the parent subdivision
    #print()
    #print(sub_division_index)
    #sub_division_index = np.argwhere(np.array(sub_division_map)==-1).flatten()
    #print(sub_division_index)
    start_time = time.time()

    for i in range(len(main_idx)):
        start = sub_division_index[main_idx[i]]
        if main_idx[i] == sub_division_index[-1]:
            end = None
        else:
            end = sub_division_index[main_idx[i]+1]
        if len(sub_division_map[start+1:end]) > 0 and main_idx[i] > 0:
            if alt_idx[i] > -1:
                alt_start = sub_division_index[alt_idx[i]]
                if alt_idx[i] == sub_division_index[-1]:
                    alt_end = None
                else:
                    alt_end = sub_division_index[alt_idx[i]+1]
                tmp = sub_division_map[alt_start+1:alt_end]
                data[tmp,28] = (data[tmp,28]/data[alt_idx[i],28])*alt_scale[i][optimum]
        data[main_idx[i],25] = RR[i][optimum]
        data[main_idx[i],27] = RL[i][optimum]
        data[main_idx[i],22] = flows[i]
        data[main_idx[i],23] = bifs[i*2][optimum]
        data[main_idx[i],24] = bifs[i*2+1][optimum]
        if main_idx[i] > 0:
            data[main_idx[i],28] = main_scale[i][optimum]
            if alt_idx[i] > -2:
                data[alt_idx[i],28] = alt_scale[i][optimum]
    data[0,21] = R0[optimum]
    data[0,23] = bifs[-2][optimum]
    data[0,24] = bifs[-1][optimum]
    data[0,25] = RR[-1][optimum]
    data[0,27] = RL[-1][optimum]
    data[0,22] = flows[-1]
    data[:,21] = data[:,28]*data[0,21]
    if not nosegments:
        segment_data[:,21] = data[segment_data[:,29].astype(int),28]*data[0,21]
        total_data = np.vstack((data,segment_data))
    else:
        total_data = data

    #total_data = update_map(data,main_idx,alt_idx,main_scale,alt_scale,RR,RL,flows,bifs,R0,sub_division_index,np.array(sub_division_map),optimum,new_sister_edge,new_terminal_edge)
    main_idx = np.flip(np.sort(main_idx))
    for i in range(len(main_idx)):
        sub_division_map.insert(sub_division_index[main_idx[i]]+1,int(new_sister_edge))
        sub_division_map.insert(sub_division_index[main_idx[i]]+1,int(new_terminal_edge))
        #sub_division_index[main_idx[i]] += (len(main_idx)-1-i)*2

    #total_data,main_idx,sub_division_index,sub_division_map = update_map(data,main_idx,alt_idx,main_scale,alt_scale,RR,RL,flows,bifs,R0,sub_division_index,np.array(sub_division_map),optimum,new_sister_edge,new_terminal_edge,sub_division_map)
    #adder = 0
    #for i in range(len(sub_division_index)):
    #    sub_division_index[i] += adder
    #    if i in main_idx:
    #        adder += 2
    tree.time['add_4'].append(time.time()-start_time)
    sub_division_index = update_index(sub_division_index,main_idx)
    #tree.time['add_4'].append(time.time()-start_time)
    #print()
    #print('main_idx: {}'.format(main_idx))
    #print('manual: {}'.format(sub_division_index))
    #sub_division_index = np.argwhere(np.array(sub_division_map)==-1).flatten()
    #print('np arg: {}'.format(sub_division_index))
    #if time.time()-start > 100:
    #    val = 0
    #else:
    #    val = time.time()-start
    #tree.time['add_4'].append(time.time()-start_time)
    return total_data,sub_division_map,sub_division_index

@nb.jit(nopython=True,cache=True,nogil=True)
def update_index(sub_division_index,main_idx):
    adder = 0
    for i in range(len(sub_division_index)):
        sub_division_index[i] += adder
        if i in main_idx:
            adder += 2
    return sub_division_index

"""
@nb.jit(nopython=True,cache=True,nogil=True)
def update_map(data,main_idx,alt_idx,main_scale,alt_scale,RR,RL,flows,bifs,R0,sub_division_index,sub_division_map,optimum,new_sister_edge,new_terminal_edge):
    for i in range(len(main_idx)):
        start = sub_division_index[main_idx[i]]
        if main_idx[i] == sub_division_index[-1]:
            end = None
        else:
            end = sub_division_index[main_idx[i]+1]
        if len(sub_division_map[start+1:end]) > 0 and main_idx[i] > 0:
            if alt_idx[i] > -1:
                alt_start = sub_division_index[alt_idx[i]]
                if alt_idx[i] == sub_division_index[-1]:
                    alt_end = None
                else:
                    alt_end = sub_division_index[alt_idx[i]+1]
                tmp = sub_division_map[alt_start+1:alt_end]
                data[tmp,28] = (data[tmp,28]/data[alt_idx[i],28])*alt_scale[i][optimum]
        data[main_idx[i],25] = RR[i][optimum]
        data[main_idx[i],27] = RL[i][optimum]
        data[main_idx[i],22] = flows[i]
        data[main_idx[i],23] = bifs[i*2][optimum]
        data[main_idx[i],24] = bifs[i*2+1][optimum]
        if main_idx[i] > 0:
            data[main_idx[i],28] = main_scale[i][optimum]
            if alt_idx[i] > -2:
                data[alt_idx[i],28] = alt_scale[i][optimum]
    data[0,21] = R0[optimum]
    data[0,23] = bifs[-2][optimum]
    data[0,24] = bifs[-1][optimum]
    data[0,25] = RR[-1][optimum]
    data[0,27] = RL[-1][optimum]
    data[0,22] = flows[-1]
    data[:,21] = data[:,28]*data[0,21]

    return data
"""



import os

brain_files = os.listdir(os.getcwd()+os.sep+'brain_testing')

if not os.path.isdir(os.getcwd()+os.sep+'brain_data'):
    os.mkdir('brain_data')

CODE ="""
# Code to generate vascular networks for Third Ventricle

import svcco
import os
import numpy as np
from pickle import dump

######################
number = 100
######################

FOLDER   = '{}'
OBJ_FILE = '{}'
NAME     = '{}_8000_vessels_'

if not os.path.isdir(os.getcwd()+os.sep+FOLDER):
    os.mkdir(os.getcwd()+os.sep+FOLDER)
    os.mkdir(os.getcwd()+os.sep+FOLDER+os.sep+'times')
    os.mkdir(os.getcwd()+os.sep+FOLDER+os.sep+'data')

FILE_TIME_PREFIX = os.getcwd()+os.sep+FOLDER+os.sep+'times'
FILE_DATA_PREFIX = os.getcwd()+os.sep+FOLDER+os.sep+'data'

s = svcco.surface()
s.load(OBJ_FILE)

s.solve()
s.build()

t = svcco.tree()
t.set_boundary(s)

t.set_root()
t.n_add(500)

EXISTING_TIMES     = len(os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+'times'))
EXISTING_DATA      = len(os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+'data'))
EXISTING_PERFUSION = len(list(filter(lambda n: 'perfusion' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+'data'))))
EXISTING_MESH      = len(list(filter(lambda n: 'mesh' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+'data'))))
EXISTING_TREE      = len(list(filter(lambda n: 'tree' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+'data'))))
EXISTING_RAW       = len(list(filter(lambda n: 'raw' in n,os.listdir(os.getcwd()+os.sep+FOLDER+os.sep+'data'))))

for i in range(number):
    t = svcco.tree()
    t.set_boundary(s)
    t.set_assumptions(convex=True)
    t.set_root()
    t.n_add(8000)
    file = open(os.getcwd() + os.sep + FOLDER +os.sep+'times'+os.sep+NAME+'{{}}.pkl'.format(i+EXISTING_TIMES),"wb")
    dump(t.time,file)
    file.close()

    id,vol = svcco.perfusion_territory(t,mesh_file=os.getcwd()+os.sep+FOLDER+os.sep+'data'+os.sep+NAME+'mesh_{{}}.vtu'.format(i+EXISTING_MESH),
                                                 tree_file=os.getcwd()+os.sep+FOLDER+os.sep+'data'+os.sep+NAME+'tree_{{}}.vtp'.format(i+EXISTING_TREE))
    file = open(os.getcwd() + os.sep + FOLDER +os.sep+'data'+os.sep+NAME+'perfusion_{{}}.pkl'.format(i+EXISTING_PERFUSION),"wb")
    dump(vol,file)
    file.close()

    np.savetxt(os.getcwd() + os.sep + FOLDER +os.sep+'data'+os.sep+NAME+'raw_{{}}.csv'.format(i+EXISTING_RAW),t.data,delimiter=',')

grid = s.pv_polydata.delaunay_3d()
surf = grid.extract_surface()
stats = [s.pv_polydata.area,s.pv_polydata.volume,surf.area,grid.volume]
file = open(os.getcwd() + os.sep + FOLDER +os.sep+'data'+os.sep+'stats.pkl',"wb")
dump(stats,file)
file.close()
"""

for file in brain_files:
    part_name = file.split('_')[-1].split('.')[0].replace(' ','_')
    full_part_path = (os.getcwd()+os.sep+'brain_testing'+os.sep+file).replace(os.sep,os.sep+os.sep)
    f = open(os.getcwd()+os.sep+'brain_data'+os.sep+part_name+'.py','w+')
    f.writelines([CODE.format(part_name,full_part_path,part_name)])
    f.close()



import numpy as np
import numba as nb
import pyvista as pv

class grid:
    def __init__(self):
        self.P = pv.Plotter(shape=(4,4))
        self.next = [0,0]
        self.size = [4,4]

    def plot(self,cyl1,cyl2,plane=None,title="None"):
        self.P.subplot(self.next[0],self.next[1])
        self.P.add_mesh(cyl1,color='g')
        self.P.add_mesh(cyl2)
        if plane is not None:
            self.P.add_mesh(plane)
        self.P.add_title(title)
        if self.next[1]+1 < self.size[1]:
            self.next[1] += 1
        else:
            self.next[0] += 1
            self.next[1] = 0

    def show(self):
        self.P.show()

@nb.jit(nopython=True)
def separating_axis(position,plane,U1,V1,W1,U2,V2,W2,U1_scale,
                    V1_scale,W1_scale,U2_scale,V2_scale,W2_scale):
    return np.abs(np.dot(position,plane)) > (np.abs(np.dot(U1*U1_scale,plane))+
                                             np.abs(np.dot(V1*V1_scale,plane))+
                                             np.abs(np.dot(W1*W1_scale,plane))+
                                             np.abs(np.dot(U2*U2_scale,plane))+
                                             np.abs(np.dot(V2*V2_scale,plane))+
                                             np.abs(np.dot(W2*W2_scale,plane)))

@nb.jit(nopython=True)
def obb(data,edge):
    #C2 = (edge[0:3] + edge[3:6]) / 2
    #C1 = (data[:,0:3] + data[:,3:6]) / 2
    """
    start = time.time()
    collision = False
    start = time.time()
    cyl_2 = vtk.vtkTubeFilter()
    line2 = vtk.vtkLineSource()
    line2.SetPoint1(edge[0],edge[1],edge[2])
    line2.SetPoint2(edge[3],edge[4],edge[5])
    cyl_2.SetInputConnection(line2.GetOutputPort())
    cyl_2.SetRadius(edge[21])
    cyl_2.SetNumberOfSides(20)
    cyl_2.CappingOn()
    for i in range(data.shape[0]):
        cyl_1 = vtk.vtkTubeFilter()
        line1 = vtk.vtkLineSource()
        line1.SetPoint1(data[i,0],data[i,1],data[i,2])
        line1.SetPoint2(data[i,3],data[i,4],data[i,5])
        cyl_1.SetInputConnection(line1.GetOutputPort())
        cyl_1.SetRadius(data[i,21])
        cyl_1.SetNumberOfSides(20)
        cyl_1.CappingOn()
        collision = vtk.vtkCollisionDetectionFilter()
        collision.SetInputConnection(0,cyl_1.GetOutputPort())
        collision.SetInputConnection(1,cyl_2.GetOutputPort())
        transform = vtk.vtkTransform()
        matrix = vtk.vtkMatrix4x4()
        collision.SetTransform(0,transform)
        collision.SetMatrix(1,matrix)
        collision.SetCollisionModeToAllContacts()
        collision.GenerateScalarsOn()
        collision.Update()
        if collision.GetNumberOfContacts() > 0:
            collision = True
            break
    #print('OBB: {}'.format(time.time()-start))
    end = time.time() - start
    #print(end)
    return collision
    """

    C1 = (edge[0:3] + edge[3:6]) / 2
    C2 = (data[:,0:3] + data[:,3:6]) / 2
    Position = C2 - C1
    #CENTER_ALL = (C2 + C1)/2
    # Basis vector construction
    edge[12:15] = ((edge[3:6] - edge[0:3]) /
                   np.linalg.norm(edge[3:6] -
                                  edge[0:3]))
    if edge[14] == -1:
        edge[6:9] = np.array([-1,0,0])
        edge[9:12] = np.array([0,-1,0])
    else:
        edge[6:9] = np.array([1-edge[12]**2/(1+edge[14]),
                              (-edge[12]*edge[13])/(1+edge[14]),
                              -edge[12]])
        edge[9:12] = np.array([(-edge[12]*edge[13])/(1+edge[14]),
                               1 - edge[13]**2/(1+edge[14]),
                               -edge[13]])
    edge[20] = np.linalg.norm(edge[3:6] - edge[0:3])
    #print(np.linalg.norm(edge[6:9]))
    #print(np.linalg.norm(edge[9:12]))
    #print(np.linalg.norm(edge[12:15]))
    result = []
    for i in range(data.shape[0]):
        #CENTER = CENTER_ALL[i]
        #G = grid()
        # U1 axis
        #cyl1 = pv.Cylinder(center=(data[i,0:3]+data[i,3:6])/2,direction=data[i,12:15],
        #                   radius=data[i,21],height=data[i,20])
        #cyl2 = pv.Cylinder(center=(edge[0:3]+edge[3:6])/2,direction=edge[12:15],
        #                   radius=edge[21],height=edge[20])
        #G.plot(cyl1,cyl2,plane=None,title="Objects")
        #plane = pv.Plane(center=CENTER,direction=edge[6:9])
        #G.plot(cyl1,cyl2,plane=plane,title="U1 axis")
        if separating_axis(Position[i,:],edge[6:9],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # V1 axis
        #plane = pv.Plane(center=CENTER,direction=edge[9:12])
        #G.plot(cyl1,cyl2,plane=plane,title="V1 axis")
        if separating_axis(Position[i,:],edge[9:12],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # W1 axis
        #plane = pv.Plane(center=CENTER,direction=edge[12:15])
        #G.plot(cyl1,cyl2,plane=plane,title="W1 axis")
        if separating_axis(Position[i,:],edge[12:15],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # U2 axis
        #plane = pv.Plane(center=CENTER,direction=data[i,6:9])
        #G.plot(cyl1,cyl2,plane=plane,title="U2 axis")
        if separating_axis(Position[i,:],data[i,6:9],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # V2 axis
        #plane = pv.Plane(center=CENTER,direction=data[i,9:12])
        #G.plot(cyl1,cyl2,plane=plane,title="V2 axis")
        if separating_axis(Position[i,:],data[i,9:12],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # W2 axis
        #plane_obj = pv.Plane(center=CENTER,direction=data[i,12:15])
        #G.plot(cyl1,cyl2,plane=plane,title="W2 axis")
        if separating_axis(Position[i,:],data[i,12:15],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # Mixed Plane Separation Tests
        # U1 X U2
        plane = np.cross(edge[6:9],data[i,6:9])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="U1 x U2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # U1 X V2
        plane = np.cross(edge[6:9],data[i,9:12])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="U1 x V2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # U1 X W2
        plane = np.cross(edge[6:9],data[i,12:15])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="U1 x W2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # V1 X U2
        plane = np.cross(edge[9:12],data[i,6:9])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="V1 x U2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # V1 X V2
        plane = np.cross(edge[9:12],data[i,9:12])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="V1 x V2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # V1 X W2
        plane = np.cross(edge[9:12],data[i,12:15])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="V1 x W2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # W1 X U2
        plane = np.cross(edge[12:15],data[i,6:9])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="W1 x U2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # W1 X V2
        plane = np.cross(edge[12:15],data[i,9:12])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="W1 x V2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        # W1 X W2
        plane = np.cross(edge[12:15],data[i,12:15])
        #plane_obj = pv.Plane(center=CENTER,direction=plane)
        #G.plot(cyl1,cyl2,plane=plane_obj,title="W1 x W2 axis")
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            result.append(False)
            #G.show()
            continue
        else:
            #print('Collision')
            #G.show()
            """
            colors = vtk.vtkNamedColors()
            background_color = 'white'
            cyl = vtk.vtkTubeFilter()
            line = vtk.vtkLineSource()
            line.SetPoint1(edge[0],edge[1],edge[2])
            line.SetPoint2(edge[3],edge[4],edge[5])
            cyl.SetInputConnection(line.GetOutputPort())
            cyl.SetRadius(edge[21])
            cyl.SetNumberOfSides(100)
            #models.append(cyl)
            mapper = vtk.vtkPolyDataMapper()
            actor1  = vtk.vtkActor()
            mapper.SetInputConnection(cyl.GetOutputPort())
            actor1.SetMapper(mapper)
            actor1.GetProperty().SetColor(colors.GetColor3d('green'))
            #actors.append(actor)
            colors = vtk.vtkNamedColors()
            cyl = vtk.vtkTubeFilter()
            line = vtk.vtkLineSource()
            line.SetPoint1(data[i,0],data[i,1],data[i,2])
            line.SetPoint2(data[i,3],data[i,4],data[i,5])
            cyl.SetInputConnection(line.GetOutputPort())
            cyl.SetRadius(data[i,21])
            cyl.SetNumberOfSides(100)
            #models.append(cyl)
            mapper = vtk.vtkPolyDataMapper()
            actor2  = vtk.vtkActor()
            mapper.SetInputConnection(cyl.GetOutputPort())
            actor2.SetMapper(mapper)
            actor2.GetProperty().SetColor(colors.GetColor3d('green'))
            #actors.append(actor)
            renderer = vtk.vtkRenderer()
            renderer.SetBackground(colors.GetColor3d(background_color))

            render_window = vtk.vtkRenderWindow()
            render_window.AddRenderer(renderer)
            render_window.SetWindowName('SimVascular Vessel Collision')

            interactor = vtk.vtkRenderWindowInteractor()
            interactor.SetRenderWindow(render_window)
            renderer.AddActor(actor1)
            renderer.AddActor(actor2)
            render_window.Render()
            interactor.Start()
            """
            return True

    return False

@nb.jit(nopython=True)
def obbc(data,edge):
    C1 = (edge[0:3] + edge[3:6]) / 2
    C2 = (data[:,0:3] + data[:,3:6]) / 2
    Position = C2 - C1
    # Basis vector construction
    edge[12:15] = ((edge[3:6] - edge[0:3]) /
                   np.linalg.norm(edge[3:6] -
                                  edge[0:3]))
    if edge[14] == -1:
        edge[6:9] = np.array([-1,0,0])
        edge[9:12] = np.array([0,-1,0])
    else:
        edge[6:9] = np.array([1-edge[12]**2/(1+edge[14]),
                              (-edge[12]*edge[13])/(1+edge[14]),
                              -edge[12]])
        edge[9:12] = np.array([(-edge[12]*edge[13])/(1+edge[14]),
                               1 - edge[13]**2/(1+edge[14]),
                               -edge[13]])
    edge[20] = np.linalg.norm(edge[3:6] - edge[0:3])
    result = []
    for i in range(data.shape[0]):
        # U1 axis
        #cyl1 = pv.Cylinder(center=(data[i,0:3]+data[i,3:6])/2,direction=data[i,12:15],
        #                   radius=data[i,21],height=data[i,20])
        #cyl2 = pv.Cylinder(center=(edge[0:3]+edge[3:6])/2,direction=edge[12:15],
        #                   radius=edge[21],height=edge[20])
        if separating_axis(Position[i,:],edge[6:9],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # V1 axis
        if separating_axis(Position[i,:],edge[9:12],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # W1 axis
        if separating_axis(Position[i,:],edge[12:15],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # U2 axis
        if separating_axis(Position[i,:],data[i,6:9],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # V2 axis
        if separating_axis(Position[i,:],data[i,9:12],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # W2 axis
        if separating_axis(Position[i,:],data[i,12:15],edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # Mixed Plane Separation Tests
        # U1 X U2
        plane = np.cross(edge[6:9],data[i,6:9])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # U1 X V2
        plane = np.cross(edge[6:9],data[i,9:12])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # U1 X W2
        plane = np.cross(edge[6:9],data[i,12:15])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # V1 X U2
        plane = np.cross(edge[9:12],data[i,6:9])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # V1 X V2
        plane = np.cross(edge[9:12],data[i,9:12])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # V1 X W2
        plane = np.cross(edge[9:12],data[i,12:15])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # W1 X U2
        plane = np.cross(edge[12:15],data[i,6:9])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # W1 X V2
        plane = np.cross(edge[12:15],data[i,9:12])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        # W1 X W2
        plane = np.cross(edge[12:15],data[i,12:15])
        if separating_axis(Position[i,:],plane,edge[6:9],edge[9:12],edge[12:15],
                           data[i,6:9],data[i,9:12],data[i,12:15],edge[21],edge[21],
                           edge[20]/2,data[i,21],data[i,21],data[i,20]/2):
            continue
        else:
            result.append(i)
    return result



import numba as nb
import numpy as np

@nb.jit(nopython=True,cache=True,nogil=True)
def check(data,point_set,threshold):
    centers = (data[:,0:3] + data[:,3:6])/2
    for i in range(point_set.shape[0]):
        distances = np.sqrt(np.sum(np.square(centers-point_set[i,:]),axis=1))
        if np.all(distances > threshold):
            return True, point_set[i,:]
    return False,None

@nb.jit(nopython=True,cache=True,nogil=True)
def check_point(data,point,threshold):
    centers = (data[:,0:3] + data[:,3:6])/2
    distances = np.sqrt(np.sum(np.square(centers-point),axis=1))
    if np.all(distances > threshold):
        return True
    return False


"""
Code to evaluate computational cost for bifurcation optimization
"""
import svcco
import numpy as np
from time import time, perf_counter
from copy import deepcopy
import matplotlib.pyplot as plt

#####################################################
# CODE FOR BUILDING SVCCO Tree
#####################################################
import pyvista as pv

cube = pv.Cube().triangulate().subdivide(3)

s = svcco.surface()
s.set_data(10*cube.points,normals=cube.point_normals)
s.solve()
s.build()

t = svcco.tree()
t.set_boundary(s)
t.set_root()
t.convex = True
t.n_add(10)
#####################################################

def test_naive(t):
    rng_points,_ = t.boundary.pick(size=len(t.boundary.tet_verts),homogeneous=True,replacement=False)
    rng_points = rng_points.tolist()
    point = np.array(rng_points.pop(0))
    vessel, line_distances = svcco.branch_addition.close.close_exact(t.data,point)
    vessel = vessel[0]
    sampling=20
    points   = svcco.branch_addition.triangle.get_local_points(t.data,vessel,point.flatten(),sampling,t.clamped_root)
    mu                 = t.parameters['mu']
    lam                = t.parameters['lambda']
    gamma              = t.parameters['gamma']
    nu                 = t.parameters['nu']
    Qterm              = t.parameters['Qterm']
    Pperm              = t.parameters['Pperm']
    Pterm              = t.parameters['Pterm']
    data = deepcopy(t.data)
    start = perf_counter()
    svcco.branch_addition.finite_difference.finite_difference_naive(data,points,point.flatten(),vessel,gamma,nu,Qterm,Pperm,Pterm)
    elapsed = perf_counter() - start
    return elapsed

def test_binding(t):
    rng_points,_ = t.boundary.pick(size=len(t.boundary.tet_verts),homogeneous=True,replacement=False)
    rng_points = rng_points.tolist()
    point = np.array(rng_points.pop(0))
    vessel, line_distances = svcco.branch_addition.close.close_exact(t.data,point)
    vessel = vessel[0]
    sampling=20
    points = svcco.branch_addition.triangle.get_local_points(t.data,vessel,point.flatten(),sampling,t.clamped_root)
    mu                 = t.parameters['mu']
    lam                = t.parameters['lambda']
    gamma              = t.parameters['gamma']
    nu                 = t.parameters['nu']
    Qterm              = t.parameters['Qterm']
    Pperm              = t.parameters['Pperm']
    Pterm              = t.parameters['Pterm']
    data = deepcopy(t.data)
    data = np.vstack((data,np.zeros((2,data.shape[1]))))
    start = perf_counter()
    svcco.branch_addition.local_func_v7.fast_local_function(data,points,point.flatten(),vessel,gamma,nu,Qterm,Pperm,Pterm)
    elapsed = perf_counter() - start
    return elapsed

def test_bif(n_steps=100,step=1):
    t = svcco.tree()
    t.set_boundary(s)
    t.set_root()
    t.convex = True
    t.n_add(1)
    SIZE       = []
    TIME_NAIVE = []
    TIME_BIND  = []
    t.rng_points,_ = t.boundary.pick(size=40*n_steps,homogeneous=True)
    t.rng_points = t.rng_points.tolist()
    for i in range(n_steps):
        SIZE.append(t.parameters['edge_num'])
        #start = perf_counter()
        #t.add(-1,0)
        #elapsed = perf_counter() - start
        TIME_BIND.append(test_binding(t))
        TIME_NAIVE.append(test_naive(t))
        t.add(-1,0)
        #TIME_BIND.append(test_binding(t))
        #t.n_add(step)
    #t.show()
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    #SIZE.insert(0,1)
    #TIME_BIND.insert(0,TIME_BIND[0]/SIZE[1])
    #TIME_NAIVE.insert(0,TIME_NAIVE[0]/SIZE[1])
    SIZE = np.array(SIZE)
    TIME_BIND = np.array(TIME_BIND)
    TIME_NAIVE = np.array(TIME_NAIVE)
    #ax.loglog(SIZE,(TIME_BIND[0]/SIZE[0])*SIZE**2*np.log(SIZE),label="Literature Implementation")
    ax.loglog(SIZE,np.cumsum(TIME_BIND),label="Partial Binding")
    ax.loglog(SIZE,np.cumsum(TIME_NAIVE),label="Naive Scaling")
    ax.loglog(SIZE,(TIME_BIND[0]/SIZE[0])*SIZE*np.log(SIZE),label="Guy Floor Scaling")
    #ax.loglog(SIZE,(TIME_BIND[0]/SIZE[0])*np.log(SIZE),label="Log(N) Scaling")
    ax.loglog(SIZE,(TIME_BIND[0]/SIZE[0])*SIZE**2,label="Quadratic Scaling")
    #ax.semilogy(np.log(SIZE),(TIME_NAIVE[0]/3)*SIZE**(4/3),label="Predicted Scaling")
    ax.legend(loc='upper left')
    plt.show()



# Patch functions
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits import mplot3d
import numpy as np
from matplotlib.ticker import MultipleLocator, FixedLocator
import matplotlib.ticker as mticker

def gen_fig(h=0,n_lines=9):
    r = np.linspace(-2,2,num=1000)
    fig,ax = plt.subplots()
    c = np.arange(1, n_lines + 1)
    norm = mpl.colors.Normalize(vmin=c.min(), vmax=c.max())
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.Blues)
    cmap.set_array([])
    for q in range(n_lines):
        w = (1-np.tanh(abs(r)))/(abs(r)+h)**q
        ax.plot(r,w,c=cmap.to_rgba(q+1),label="q={}".format(q))
    ax.set_xlabel(r'Normalized Distance, $r$')
    ax.set_ylabel('Weight')
    ax.set_yscale('log')
    fig.colorbar(cmap, ticks=c,label=r"Patch Degree, $q$")
    name = "Patch_functions_q{}_h{}.svg".format(n_lines,h)
    fig.savefig(name,format="svg")
    plt.show()

def log_tick_formatter(val, pos=None):
    return f"10$^{{{int(val)}}}$"

def gen_3d_fig(h=0,q=1,n_lines=9):
    x_interval = (-1, 1)
    y_interval = (-1, 1)
    x_points = np.linspace(x_interval[0], x_interval[1], 1000)
    y_points = np.linspace(y_interval[0], y_interval[1], 1000)
    X, Y = np.meshgrid(x_points, y_points)
    #r = (X**2 + Y**2)**(1/2)
    #Z = (1-np.tanh(r))/(r+h)**q
    c = np.arange(1, n_lines + 1)
    norm = mpl.colors.Normalize(vmin=c.min(), vmax=c.max())
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.Blues)
    cmap.set_array([])
    fig = plt.figure()
    ax = plt.axes(projection="3d")
    #for q in range(n_lines):
    r = (X**2 + Y**2)**(1/2)
    Z = (1-np.tanh(r))/(r+h)**q
    ax.plot_surface(X,Y,np.log10(Z),color=cmap.to_rgba(q+1),label="q={}".format(q))
    ax.grid(False)
    Zmin = np.where(Z > 0, Z, np.inf).min()
    #print(np.ceil(np.log10(Z.max())))
    #print(np.floor(np.log10(Zmin))-1)
    ax.set_zlim3d([np.floor(np.log10(Zmin)),np.ceil(np.log10(Z.max()))])
    ax.zaxis.set_major_formatter(mticker.FuncFormatter(log_tick_formatter))
    ax.zaxis.set_major_locator(MultipleLocator(1))
    zminorticks = []
    zaxmin, zaxmax = ax.get_zlim()
    for zorder in np.arange(np.floor(zaxmin),np.ceil(zaxmax)):
        zminorticks.extend(np.log10(np.linspace(2,9,8)) + zorder)
    ax.zaxis.set_minor_locator(FixedLocator(zminorticks))
    ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    name = "3D_Patch_functions_q{}_h{}.svg".format(q,h)
    fig.savefig(name,format="svg")
    plt.show()

gen_fig(h=0)
gen_fig(h=1)
gen_3d_fig(h=0,q=1)
gen_3d_fig(h=0,q=8)
gen_3d_fig(h=1,q=1)
gen_3d_fig(h=1,q=8)



import svcco
import pyvista as pv
import numpy as np

inlet_point   = np.array([0,0,-0.5])
outlet_point  = np.array([0,0,0.5])
inlet_normal  = np.array([0,0,1])
outlet_normal = np.array([0,0,-1])
cube          = pv.Cube().triangulate().subdivide(3)

s = svcco.surface()
s.set_data(cube.points,cube.point_normals)
s.solve()
s.build()


t = svcco.forest(boundary=s,number_of_networks=1,trees_per_network=[2],convex=True,
                 start_points=[[inlet_point,outlet_point]],directions=[[inlet_normal,outlet_normal]],
                 root_lengths_low=[[0.25,0.25]],root_lengths_high=[[0.5,0.5]])

#t.set_roots()

#p = t.show()
#p.show()

t.networks[0][0].set_parameters(Pperm=5.1,Pterm=5,edge_num=1,Qterm=(0.25/21))
t.networks[0][1].set_parameters(Pperm=5.1,Pterm=5,edge_num=1,Qterm=(0.25/21))

t.set_roots()
t.add(30)
t.connect(curve_sample_size_min=11,curve_sample_size_max=19)
p = t.forest_copy.show()
p.show()
#print(t.networks[0][0].data[0,20])

#p = t.show()
#p.show()

t.add(1)
